# -*- coding: utf-8 -*-
"""Provide class to manage a simulation."""

import smuthi.linearsystem.linear_system as lsys
import smuthi.fields as flds
import sys
import os
import datetime
import time
import shutil
import pickle
import numpy as np
import smuthi
import warnings
import smuthi.utility.logging as log


class Simulation:
    """Central class to manage a simulation.

    Args:
        layer_system (smuthi.layers.LayerSystem):               stratified medium
        particle_list (list):                                   list of smuthi.particles.Particle objects
        initial_field (smuthi.initial_field.InitialField):      initial field object
        k_parallel (numpy.ndarray or str):              in-plane wavenumber for Sommerfeld integrals and field
                                                        expansions. if 'default', use
                                                        smuthi.fields.default_Sommerfeld_k_parallel_array
        angular_resolution (float):                     If provided, angular arrays are generated with this angular
                                                        resolution over the default angular range
        neff_waypoints (list or ndarray):               Used to set default k_parallel arrays.
                                                        Corner points through which the contour runs
                                                        This quantity is dimensionless (effective
                                                        refractive index, will be multiplied by vacuum
                                                        wavenumber)
                                                        :ref:`MultipoleCutOffAnchor`
                                                        If not provided, reasonable waypoints are estimated.
        neff_imag (float):                              Used to set default k_parallel arrays.
                                                        Extent of the contour into the negative imaginary direction
                                                        (in terms of effective refractive index, n_eff=kappa/omega).
                                                        Only needed when no neff_waypoints are provided
        neff_max (float):                               Used to set default k_parallel arrays.
                                                        Truncation value of contour (in terms of effective refractive
                                                        index). Only needed when no neff_waypoints are
                                                        provided
        neff_max_offset (float):                        Used to set default k_parallel arrays.
                                                        Use the last estimated singularity location plus this value
                                                        (in terms of effective refractive index). Default=1
                                                        Only needed when no `neff_waypoints` are provided
                                                        and if no value for `neff_max` is specified.
        neff_resolution(float):                         Used to set default k_parallel arrays.
                                                        Resolution of contour, again in terms of effective refractive
                                                        index
        neff_minimal_branchpoint_distance (float):      Used to set default k_parallel arrays.
                                                        Minimal distance that contour points shall have from
                                                        branchpoint singularities (in terms of effective
                                                        refractive index). This is only relevant if not deflected
                                                        into imaginary. Default: One fifth of neff_resolution
        overwrite_default_contours (bool):              If true (default), the default contours are written even if
                                                        they have already been defined before
        solver_type (str):                      What solver type to use?
                                                Options: 'LU' for LU factorization, 'gmres' for GMRES iterative solver
        coupling_matrix_lookup_resolution (float or None): If type float, compute particle coupling by interpolation of
                                                           a lookup table with that spacial resolution. If None
                                                           (default), don't use a lookup table but compute the coupling
                                                           directly. This is more suitable for a small particle number.
        coupling_matrix_interpolator_kind (str):  Set to 'linear' (default) or 'cubic' interpolation of the lookup table.
        store_coupling_matrix (bool):           If True (default), the coupling matrix is stored. Otherwise it is
                                                recomputed on the fly during each iteration of the solver.
        length_unit (str):      what is the physical length unit? has no influence on the computations
        input_file (str):       path and filename of input file (for logging purposes)
        output_dir (str):       path to folder where to export data
        save_after_run(bool):   if true, the simulation object is exported to disc when over
        log_to_file(bool):      if true, the simulation log will be written to a log file
        log_to_terminal(bool):  if true, the simulation progress will be displayed in the terminal
        check_circumscribing_spheres(bool):  if true, check all particles for overlapping circumscribing spheres
                                             and print a warning if detected
        do_sanity_check (bool):              if true (default), check numerical input for some flaws. Warning: A passing
                                             sanity check does not guarantee correct numerical settings. For many
                                             particles, the sanity check might take some time and/or occupy large memory.
        periodicity (tuple):                 tuple (a1, a2) specifying two 3-dimensional lattice vectors in Carthesian coordinates
                                             with a1, a2 (numpy.ndarrays)
        ewald_sum_separation_parameter (float):     Used to separate the real and reciprocal lattice sums to evaluate
                                                    particle coupling in periodic lattices.
        number_of_threads_periodic (int or str):     sets the number of threats used in a simulation with periodic particle arrangements
                                                    if 'default', all available CPU cores are used 
                                                    if negative, all but number_of_threads_periodic are used
        use_pvwf_coupling (bool):              If set to True, plane wave coupling is used to calculate
                                               the direct. Currently only possible in combination with direct solver strategy.
        pvwf_coupling_neff_max (float):        Truncation neff for the integration contour of the PVWF coupling integral
        pvwf_coupling_neff_resolution (float): Discretization of the neff integral for PVWF coupling
    """
    def __init__(self,
                 layer_system=None,
                 particle_list=None,
                 initial_field=None,
                 k_parallel='default',
                 angular_resolution=np.pi/360,
                 neff_waypoints=None,
                 neff_imag=1e-2,
                 neff_max=None,
                 neff_max_offset=1,
                 neff_resolution=1e-2,
                 neff_minimal_branchpoint_distance=None,
                 overwrite_default_contours=True,
                 solver_type='LU',
                 solver_tolerance=1e-4,
                 store_coupling_matrix=True,
                 coupling_matrix_lookup_resolution=None,
                 coupling_matrix_interpolator_kind='linear',
                 length_unit='length unit',
                 input_file=None,
                 output_dir='smuthi_output',
                 save_after_run=False,
                 log_to_file=False,
                 log_to_terminal=True,
                 check_circumscribing_spheres=True,
                 do_sanity_check=True,
                 periodicity=None,
                 ewald_sum_separation_parameter='default',
                 number_of_threads_periodic='default',
                 use_pvwf_coupling=False,
                 pvwf_coupling_neff_max=None,
                 pvwf_coupling_neff_resolution=1e-2):

        # initialize attributes
        self.layer_system = layer_system
        self.particle_list = particle_list
        self.initial_field = initial_field
        self.k_parallel = k_parallel
        self.angular_resolution = angular_resolution
        self.neff_waypoints = neff_waypoints
        self.neff_imag = neff_imag
        self.neff_max = neff_max
        self.neff_max_offset = neff_max_offset
        self.neff_resolution = neff_resolution
        self.neff_minimal_branchpoint_distance = neff_minimal_branchpoint_distance
        self.overwrite_default_contours = overwrite_default_contours
        self.solver_type = solver_type
        self.solver_tolerance = solver_tolerance
        self.store_coupling_matrix = store_coupling_matrix
        self.coupling_matrix_lookup_resolution = coupling_matrix_lookup_resolution
        self.coupling_matrix_interpolator_kind = coupling_matrix_interpolator_kind
        self.post_processing = None
        self.length_unit = length_unit
        self.save_after_run = save_after_run
        self.check_circumscribing_spheres = check_circumscribing_spheres
        self.do_sanity_check = do_sanity_check
        self.periodicity = periodicity
        self.ewald_sum_separation_parameter = ewald_sum_separation_parameter
        self.number_of_threads_periodic = number_of_threads_periodic
        self.use_pvwf_coupling = use_pvwf_coupling
        self.pvwf_coupling_neff_max = pvwf_coupling_neff_max
        self.pvwf_coupling_neff_resolution = pvwf_coupling_neff_resolution

        # output
        timestamp = '{:%Y%m%d%H%M%S}'.format(datetime.datetime.now())
        self.output_dir = output_dir + '/' + timestamp
        self.log_to_terminal = log_to_terminal
        self.log_to_file = log_to_file
        self.log_filename = self.output_dir + '/' + 'smuthi.log'
        self.old_log = sys.stdout
        

        if input_file is not None and log_to_file:
            shutil.copyfile(input_file, self.output_dir + '/input.dat')

    def set_logging(self, log_to_terminal=None, log_to_file=None, log_filename=None):
        """Update logging behavior.

        Args:
            log_to_terminal (logical):  If true, print output to console.
            log_to_file (logical):      If true, print output to file
            log_filename (char):        If `log_to_file` is true, print output to a file with that name in the output
                                        directory. If the file already exists, it will be appended.
        """
        if log_to_terminal is not None:
            self.log_to_terminal = log_to_terminal
        if log_to_file is not None:
            self.log_to_file = log_to_file
        if log_filename is not None:
            self.log_filename = log_filename

        if not os.path.exists(self.output_dir) and self.log_to_file:
            os.makedirs(self.output_dir)
        sys.stdout = log.Logger(log_filename=self.log_filename,
                            log_to_file=self.log_to_file,
                            log_to_terminal=self.log_to_terminal)

    def __getstate__(self):
        """Return state values to be pickled."""
        return (self.layer_system, self.particle_list, self.initial_field, self.k_parallel, self.solver_type,
                self.solver_tolerance, self.store_coupling_matrix, self.coupling_matrix_lookup_resolution,
                self.coupling_matrix_interpolator_kind, self.post_processing, self.length_unit, self.save_after_run,
                self.periodicity, self.ewald_sum_separation_parameter, self.number_of_threads_periodic,
                flds.default_Sommerfeld_k_parallel_array, flds.default_initial_field_k_parallel_array,
                flds.default_polar_angles, flds.default_azimuthal_angles)

    def __setstate__(self, state):
        """Restore state from the unpickled state values."""
        (self.layer_system, self.particle_list, self.initial_field, self.k_parallel, self.solver_type,
         self.solver_tolerance, self.store_coupling_matrix, self.coupling_matrix_lookup_resolution,
         self.coupling_matrix_interpolator_kind, self.post_processing, self.length_unit, self.save_after_run,
         self.periodicity, self.ewald_sum_separation_parameter, self.number_of_threads_periodic,
         flds.default_Sommerfeld_k_parallel_array, flds.default_initial_field_k_parallel_array,
         flds.default_polar_angles, flds.default_azimuthal_angles) = state

    def print_simulation_header(self):
        smuthi.print_smuthi_header()
        sys.stdout.write("Starting simulation.\n")
        sys.stdout.flush()

    def save(self, filename=None):
        """Export simulation object to disc.

        Args:
            filename (str): path and file name where to store data
        """
        if filename is None:
            filename = self.output_dir + '/simulation.p'
        with open(filename, 'wb') as fn:
            pickle.dump(self, fn, -1)

    def sanity_check(self):
        """Check contour parameters for obvious problems"""
        maxrho = self.largest_lateral_distance()
        k = flds.angular_frequency(self.initial_field.vacuum_wavelength)
        thresh = 2 / (maxrho * k)

        # check neff_resol
        if self.neff_resolution is not None and type(self.k_parallel) is str and self.k_parallel == "default":
            if self.neff_resolution > thresh:
                sys.stdout.write("Warning: Sanity check for neff_resolution failed!\n")
                sys.stdout.write("         neff_resolution = %.3e, max_rho = %.3e. k = %.3e.\n"%(self.neff_resolution,maxrho,k))
                sys.stdout.write("         It is recommended to select neff_resolution < 2 (k max_rho) = %.3e\n"%thresh)
                sys.stdout.flush()

        # check neff_imag
        if self.neff_imag is not None and type(self.k_parallel) is str and self.k_parallel == "default" and self.neff_waypoints is None:
            if self.neff_imag > thresh:
                sys.stdout.write("Warning: Sanity check for neff_imag failed!\n")
                sys.stdout.write("         neff_imag = %.3e, max_rho = %.3e. k = %.3e.\n"%(self.neff_imag,maxrho,k))
                sys.stdout.write("         It is recommended to select neff_imag < 2 (k max_rho) = %.3e\n"%thresh)
                sys.stdout.flush()

    def largest_lateral_distance(self):
        """Compute the largest lateral distance between any two particles"""
        particle_x_array = np.array([particle.position[0] for particle in self.particle_list])
        particle_y_array = np.array([particle.position[1] for particle in self.particle_list])
        rho_squared = ((particle_x_array[:, None] - particle_x_array[None, :]) ** 2
                       + (particle_y_array[:, None] - particle_y_array[None, :]) ** 2)
        return np.sqrt(np.max(rho_squared))

    def initialize_linear_system(self):
        if self.use_pvwf_coupling:
            pvwf_kpar = flds.reasonable_Sommerfeld_kpar_contour(vacuum_wavelength=self.initial_field.vacuum_wavelength,
                                                                layer_refractive_indices=self.layer_system.refractive_indices,
                                                                neff_waypoints=[0, self.pvwf_coupling_neff_max],
                                                                neff_resolution=self.pvwf_coupling_neff_resolution)
        else:
            pvwf_kpar = None

        self.linear_system = lsys.LinearSystem(particle_list=self.particle_list,
                                               initial_field=self.initial_field,
                                               layer_system=self.layer_system,
                                               k_parallel=self.k_parallel,
                                               solver_type=self.solver_type,
                                               solver_tolerance=self.solver_tolerance,
                                               store_coupling_matrix=self.store_coupling_matrix,
                                               coupling_matrix_lookup_resolution=self.coupling_matrix_lookup_resolution,
                                               interpolator_kind=self.coupling_matrix_interpolator_kind,
                                               periodicity=self.periodicity,
                                               ewald_sum_separation_parameter=self.ewald_sum_separation_parameter,
                                               number_of_threads_periodic=self.number_of_threads_periodic,
                                               use_pvwf_coupling=self.use_pvwf_coupling,
                                               pvwf_coupling_k_parallel=pvwf_kpar
                                               )

    def circumscribing_spheres_disjoint(self):
        """Check if all circumscribing spheres are disjoint"""
        from scipy.spatial.distance import pdist, cdist, squareform
        pos = np.array([p.position for p in self.particle_list])
        csr = np.array([p.circumscribing_sphere_radius() for p in self.particle_list])
        try: # fully vectorized but memory expensive, finds all overlapping pairs
            distmatrix = squareform(pdist(pos))
            r1, r2 = np.meshgrid(csr, csr)
            overlap = np.triu(distmatrix < r1+r2, k=1)
            pidx = np.where(overlap)
            s = 's' if np.count_nonzero(overlap) > 1 else '' # pluralize message string
            msg = 'The circumscribing sphere of particle' + s
            for idx in pidx[0]:
                msg += " %i"%idx
            msg += ' overlaps with that of particle'+ s
            for idx in pidx[1]:
                msg += " %i"%idx
            msg += '.\n'
        except MemoryError: # less vectorized, more time consuming, stops at first overlap detected
            for i in range(len(self.particle_list)):
                dists = cdist(np.expand_dims(pos[i,:],0), pos[i+1:,])
                overlap = np.squeeze(dists < csr[i] + csr[i+1:])
                if overlap.any():
                    pidx = np.where(overlap)
                    msg = 'Found overlap between circumscribing sphere of particle %i and particle %i.\n'%(i,i+1+pidx[0][0])
                    break
        if overlap.any():
            sys.stdout.write(msg)
            sys.stdout.flush()
        return not overlap.any()

    def set_default_Sommerfeld_contour(self):
        """Set the default Sommerfeld k_parallel array"""

        flds.default_Sommerfeld_k_parallel_array = flds.reasonable_Sommerfeld_kpar_contour(
            vacuum_wavelength=self.initial_field.vacuum_wavelength,
            neff_waypoints=self.neff_waypoints,
            layer_refractive_indices=self.layer_system.refractive_indices,
            neff_imag=self.neff_imag,
            neff_max=self.neff_max,
            neff_max_offset=self.neff_max_offset,
            neff_resolution=self.neff_resolution,
            neff_minimal_branchpoint_distance=self.neff_minimal_branchpoint_distance)

    def set_default_angles(self):
        """Set the default polar and azimuthal angular arrays for pre-processing (i.e., initial field expansion)"""

        flds.default_azimuthal_angles, flds.default_polar_angles = \
        flds.angular_arrays(angular_resolution=self.angular_resolution)

    def set_default_initial_field_contour(self):
        """Set the default initial field k_parallel array"""

        if type(self.initial_field).__name__ == 'GaussianBeam':
            # in that case use only wavenumbers that propagate in the originating layer
            if self.initial_field.polar_angle <= np.pi / 2:
                neff_max = self.layer_system.refractive_indices[0].real
            else:
                neff_max = self.layer_system.refractive_indices[-1].real

            flds.default_initial_field_k_parallel_array = flds.reasonable_Sommerfeld_kpar_contour(
                vacuum_wavelength=self.initial_field.vacuum_wavelength,
                layer_refractive_indices=self.layer_system.refractive_indices,
                neff_imag=0,
                neff_max=neff_max,
                neff_resolution=self.neff_resolution,
                neff_minimal_branchpoint_distance=self.neff_minimal_branchpoint_distance)
        else:
            # case of dipoles etc ...
            # use a similar contour as for Sommerfeld integrals
            flds.default_initial_field_k_parallel_array = flds.reasonable_Sommerfeld_kpar_contour(
                vacuum_wavelength=self.initial_field.vacuum_wavelength,
                neff_waypoints=self.neff_waypoints,
                layer_refractive_indices=self.layer_system.refractive_indices,
                neff_imag=self.neff_imag,
                neff_max=self.neff_max,
                neff_max_offset=self.neff_max_offset,
                neff_resolution=self.neff_resolution,
                neff_minimal_branchpoint_distance=self.neff_minimal_branchpoint_distance)

    def set_default_contours(self):
        """Set the default initial field k_parallel array and the default Sommerfeld k_parallel array"""
        self.set_default_initial_field_contour()
        self.set_default_Sommerfeld_contour()

    def run(self):
        self.set_logging()
        """Start the simulation."""
        self.print_simulation_header()

        # check for circumscribing sphere collisions.
        if self.check_circumscribing_spheres and len(self.particle_list) > 1 \
                and not self.circumscribing_spheres_disjoint():
            warnings.warn("Particles with circumscribing spheres detected.")

        # run sanity check
        if self.do_sanity_check and len(self.particle_list) > 1:
            self.sanity_check()

        # set default angular arrays
        if flds.default_azimuthal_angles is None or flds.default_polar_angles is None:
            self.set_default_angles()

        # check if default contours exists, otherwise set them
        if flds.default_Sommerfeld_k_parallel_array is None or self.overwrite_default_contours:
            self.set_default_Sommerfeld_contour()

        if flds.default_initial_field_k_parallel_array is None or self.overwrite_default_contours:
            self.set_default_initial_field_contour()

        # prepare and solve linear system
        start = time.time()
        self.initialize_linear_system()
        self.linear_system.prepare()
        end = time.time()
        preparation_time = end - start

        start = time.time()
        self.linear_system.solve()
        end = time.time()
        solution_time = end - start

        # post processing
        start = time.time()
        if self.post_processing:
            self.post_processing.run(self)
        end = time.time()
        postprocessing_time = end - start

        if self.save_after_run:
            self.save(self.output_dir + '/simulation.p')

        sys.stdout.write('\n')
        sys.stdout.flush()
        sys.stdout = self.old_log

        return preparation_time, solution_time, postprocessing_time



"""This module contains post processing functions to evaluate the scattered
electromagnetic field of periodic particle arrangements."""

import smuthi.fields as flds
import smuthi.fields.expansions as fldex
from smuthi.fields.transformations import transformation_coefficients_vwf
from smuthi.postprocessing.far_field import FarField
import smuthi.utility.cuda as cu
try:
    import pycuda.autoinit
    import pycuda.driver as drv
    from pycuda import gpuarray
    from pycuda.compiler import SourceModule
    import pycuda.cumath
except:
    pass
import smuthi.periodicboundaries.particle_coupling as pbcoup
import smuthi.periodicboundaries.expansions_cuda as cu_src
import numpy as np
import copy
from tqdm import tqdm
import sys


def periodic_swe_to_pwe_conversion(initial_field, layer_system, particle, a1, a2):
    """ Converts the scattered field's SphericalWaveExpansion of the arbitrary central particle within
        a periodic extend of particles into a PlaneWaveExpansion object and translates it to
        the scattering layer's anchor point.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        layer_system (smuthi.layer.LayerSystem):            stratified medium
        particle (smuthi.particles.Particle):               particle (periodic arrangement) emitting scattered field
        a1 (numpy.ndarray):                                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                                 lattice vector 2 in Carthesian coordinates 
    Returns:
        pwe_up (smuthi.fields.expansions.PlaneWaveExpansion):       upwards propagating plane wave expansion of the
                                                                    scattered field of a periodic particle arrangement 
        pwe_down (smuthi.fields.expansions.PlaneWaveExpansion):     downwards propagating plane wave expansion of the
                                                                    scattered field of a periodic particle arrangement
    """
    
    # prepare kparallel and alpha discretization of the PWE
    i_swe = layer_system.layer_number(particle.position[2]) 
    if initial_field.polar_angle < np.pi:
        pfe = initial_field.piecewise_field_expansion(layer_system).expansion_list[2 * i_swe]
    else:
        pfe = initial_field.piecewise_field_expansion(layer_system).expansion_list[2 * i_swe + 1]
    k = pfe.k
    k0t = np.array([pfe.k_parallel[0] * np.cos(pfe.azimuthal_angles),
                    pfe.k_parallel[0] * np.sin(pfe.azimuthal_angles)])
    
    A = np.linalg.norm(np.cross(a1, a2))    
    k_parallel, azimuthal_angles, kpar_alpha_comb, gmax = pbcoup.discrete_scattering_angles(k, k0t, pfe.azimuthal_angles[0], a1, a2)
    reference_point = np.array(particle.scattered_field.reference_point)
    
    # prepare pwe objects    
    lower_z_up = particle.position[2]
    upper_z_up = layer_system.upper_zlimit(i_swe)      
    pwe_up = fldex.PlaneWaveExpansion(k=k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles, kind='upgoing',
                                      reference_point=reference_point, lower_z=lower_z_up, upper_z=upper_z_up)
    lower_z_down = layer_system.lower_zlimit(i_swe)
    upper_z_down = particle.position[2]
    pwe_down = fldex.PlaneWaveExpansion(k=k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles, kind='downgoing',
                                        reference_point=reference_point, lower_z=lower_z_down, upper_z=upper_z_down)
    
    # prepare kpar and alpha lookups
    alpha_vec = np.zeros((2 * gmax + 1) ** 2)
    kpar_vec = np.zeros((2 * gmax + 1) ** 2)
    k_idx = np.zeros((2 * gmax + 1) ** 2, dtype=int)
    a_idx = np.zeros((2 * gmax + 1) ** 2, dtype=int)
    s = 0
    for idx in range(len(k_parallel)):
        alpha_vec[s:s+len(kpar_alpha_comb[idx])] = azimuthal_angles[kpar_alpha_comb[idx]]
        kpar_vec[s:s+len(kpar_alpha_comb[idx])] = k_parallel[idx]
        k_idx[s:s+len(kpar_alpha_comb[idx])] = idx
        a_idx[s:s+len(kpar_alpha_comb[idx])] = kpar_alpha_comb[idx]
        s += len(kpar_alpha_comb[idx])
        
    kz_up = flds.k_z(k_parallel=kpar_vec, k=pwe_up.k)
    
    kzk = kz_up * k 
    for l in range(1, particle.l_max + 1):
        for m in range(-min(particle.m_max, l), min(particle.m_max, l) + 1):
            for tau in range(2):
                n = flds.multi_to_single_index(tau, l, m, particle.l_max, particle.m_max)
                eima = np.exp(1j * m * alpha_vec)
                for pol in range(2):
                    B_up = transformation_coefficients_vwf(tau, l, m, pol=pol, kp=kpar_vec, kz=kz_up,
                                                           pilm_list=None, taulm_list=None, dagger=False) 
                    B_down = transformation_coefficients_vwf(tau, l, m, pol=pol, kp=kpar_vec, kz=-kz_up,
                                                             pilm_list=None, taulm_list=None, dagger=False) 
                    pwe_up.coefficients[pol, k_idx, a_idx] += particle.scattered_field.coefficients[n] * eima * B_up
                    pwe_down.coefficients[pol, k_idx, a_idx] += particle.scattered_field.coefficients[n] * eima * B_down
    pwe_up.coefficients[:, k_idx, a_idx] = 2 * np.pi / (A * kzk) * pwe_up.coefficients[:, k_idx, a_idx]
    pwe_down.coefficients[:, k_idx, a_idx] = 2 * np.pi / (A * kzk) * pwe_down.coefficients[:, k_idx, a_idx]
    
    # translate the pwe objects to the reference point
    reference_point = [0, 0, layer_system.reference_z(i_swe)]
    rpwe_mn_rswe = np.array(reference_point) - np.array(particle.scattered_field.reference_point)
    
    agrid = pwe_up.azimuthal_angle_grid()
    kpgrid = pwe_up.k_parallel_grid()
    kx = kpgrid * np.cos(agrid)
    ky = kpgrid * np.sin(agrid)
    kz_up = pwe_up.k_z_grid()
    kz_down = pwe_down.k_z_grid()

    kvec_up = np.array([kx, ky, kz_up])
    kvec_down = np.array([kx, ky, kz_down])
    ejkrSiS_up = np.nan_to_num(np.exp(1j * np.tensordot(kvec_up, rpwe_mn_rswe, axes=([0], [0]))))
    ejkrSiS_down = np.exp(1j * np.tensordot(kvec_down, rpwe_mn_rswe, axes=([0], [0])))
    
    pwe_up.reference_point = reference_point
    pwe_up.coefficients = pwe_up.coefficients * ejkrSiS_up[None, :, :]
    pwe_down.reference_point = reference_point
    pwe_down.coefficients = pwe_down.coefficients * ejkrSiS_down[None, :, :]
    
    return pwe_up, pwe_down    


def periodic_scattered_field_piecewise_expansion(initial_field, particle_list, layer_system, a1, a2,
                                                 layer_numbers=None):
    """Compute a piecewise field expansion of the scattered field of a periodic particle arrangement.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        particle_list (list):                               list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):           stratified medium
        a1 (numpy.ndarray):                                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                                 lattice vector 2 in Carthesian coordinates 
        layer_numbers (list):                               if specified, append only plane wave expansions for these layers
    Returns:
        periodic scattered field as smuthi.field_expansion.PiecewiseFieldExpansion object

    """
    if layer_numbers is None:
        layer_numbers = range(layer_system.number_of_layers())

    sfld = fldex.PiecewiseFieldExpansion()
    
    # direct field ---------------------------------------------------------------------------------------------
    for particle in particle_list:
        pwe_up, pwe_down = periodic_swe_to_pwe_conversion(initial_field, layer_system, particle, a1, a2)
        pwe_up.validity_conditions.append(particle.is_outside)
        pwe_down.validity_conditions.append(particle.is_outside)
        sfld.expansion_list.append(pwe_up)
        sfld.expansion_list.append(pwe_down)

    # layer mediated scattered field --------------------------------------------------------------------------- 
    for i in layer_numbers:        
        for ip in tqdm(range(len(particle_list)), desc='Scatt. field expansion (%i)'%i, file=sys.stdout,
                                        bar_format='{l_bar}{bar}| elapsed: {elapsed} ' 'remaining: {remaining}'):
            
            pwe_up_exc = sfld.expansion_list[2 * ip]
            pwe_down_exc =  sfld.expansion_list[2 * ip + 1]
            i_sca = layer_system.layer_number(pwe_up_exc.lower_z) 
            if ip == 0:
                pwe_up, pwe_down = layer_system.response((pwe_up_exc, pwe_down_exc), i_sca, i)
            else:
                add_up, add_down = layer_system.response((pwe_up_exc, pwe_down_exc), i_sca, i)
                
                pwe_up = pwe_up + add_up
                pwe_down = pwe_down + add_down
                
            pwe_up.validity_conditions.append(particle.is_outside)
            pwe_down.validity_conditions.append(particle.is_outside)
            
        # in bottom_layer, suppress upgoing waves, and in top layer, suppress downgoing waves
        if i > 0:
            sfld.expansion_list.append(pwe_up)
        if i < layer_system.number_of_layers()-1:
            sfld.expansion_list.append(pwe_down)

    return sfld


def total_periodic_field_piecewise_expansion(initial_field, particle_list, layer_system, a1, a2,
                                             layer_numbers=None):
    """Compute a piecewise field expansion of the total field of a periodic particle arrangement.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        particle_list (list):                               list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):           stratified medium
        a1 (numpy.ndarray):                                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                                 lattice vector 2 in Carthesian coordinates 
        layer_numbers (list):                               if specified, append only plane wave expansions for these layers
    Returns:
        periodic total field as smuthi.fields.expansions.PiecewiseFieldExpansion object
    """
    if layer_numbers is None:
        layer_numbers = range(layer_system.number_of_layers())
        
    # piecewise field expansion of the initial field 
    pfe_init = initial_field.piecewise_field_expansion(layer_system)
    
    # piecewise field expansion of the periodic scattered fields
    pfe_sca = periodic_scattered_field_piecewise_expansion(initial_field=initial_field,
                            particle_list=particle_list, layer_system=layer_system,
                            a1=a1, a2=a2, layer_numbers=layer_numbers)
    
    # total field
    pfe_total = copy.deepcopy(pfe_sca)
    idx_a = np.argwhere(pfe_sca.expansion_list[0].azimuthal_angles == pfe_init.expansion_list[0].azimuthal_angles)[0][0]
    idx_kpar = np.argwhere(pfe_sca.expansion_list[0].k_parallel == pfe_init.expansion_list[0].k_parallel)[0][0]
    for i in layer_numbers:   
        exp_idx = 2 * len(particle_list) + 2 * i
        if i > 0:
            pfe_total.expansion_list[exp_idx - 1].coefficients[:, idx_kpar, idx_a] += pfe_init.expansion_list[2 * i].coefficients.reshape(2)
        if i < layer_system.number_of_layers()-1:
            pfe_total.expansion_list[exp_idx].coefficients[:, idx_kpar, idx_a] += pfe_init.expansion_list[2 * i + 1].coefficients.reshape(2)
    
    # add initial field contributions from outside
    pfe_total.expansion_list.insert(2 * len(particle_list), pfe_init.expansion_list[0])
    pfe_total.expansion_list.append(pfe_init.expansion_list[-1])
            
    return pfe_total


def transmitted_plane_wave_expansion(initial_field, particle_list, layer_system, a1, a2):
    """Compute the plane wave expansion of the total transmitted field.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        particle_list (list):                               list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):           stratified medium
        a1 (numpy.ndarray):                                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                                 lattice vector 2 in Carthesian coordinates 
    Returns:
        smuthi.fields.expansions.PlaneWaveExpansion object of the total transmitted field
    """
    if 0 <= initial_field.polar_angle < np.pi / 2: 
        i_T = layer_system.number_of_layers() - 1
        pwe_init_T = initial_field.plane_wave_expansion(layer_system, i_T)[0]
    else:
        i_T = 0
        pwe_init_T = initial_field.plane_wave_expansion(layer_system, i_T)[1]
    
    pfe_sca = periodic_scattered_field_piecewise_expansion(initial_field, particle_list, layer_system, a1, a2, layer_numbers=[i_T])
        
    # layer response of all scattered fields
    pwe_T = copy.deepcopy(pfe_sca.expansion_list[-1])
    idx_a = np.argwhere(pwe_T.azimuthal_angles == pwe_init_T.azimuthal_angles)[0][0]
    idx_kpar = np.argwhere(pwe_T.k_parallel == pwe_init_T.k_parallel)[0][0]
    
    # add the initial field contribution
    pwe_T.coefficients[:, idx_kpar, idx_a] += pwe_init_T.coefficients.reshape(2)
    
    # if particles are located in the outer layer, add the direct scattered field contributions
    i_sca = layer_system.layer_number(particle_list[0].position[2]) # all particles have to be in the same layer
    if i_sca == i_T:
        for ip in range(len(particle_list)):
            if i_T == 0:
                pwe_T = pwe_T + pfe_sca.expansion_list[2 * ip + 1]
            else:
                pwe_T = pwe_T + pfe_sca.expansion_list[2 * ip]
 
    return pwe_T


def reflected_plane_wave_expansion(initial_field, particle_list, layer_system, a1, a2):
    """Compute the plane wave expansion of the total reflected field.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        particle_list (list):                               list of smuthi.particles.Particle objects
        layer_system (smuthi.layers.LayerSystem):           stratified medium
        a1 (numpy.ndarray):                                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                                 lattice vector 2 in Carthesian coordinates 
    Returns:
        smuthi.fields.expansions.PlaneWaveExpansion object of the total reflected field
    """
    if 0 <= initial_field.polar_angle < np.pi / 2: 
        i_R = 0
        pwe_init_R = initial_field.plane_wave_expansion(layer_system, i_R)[1]
    else:
        i_R = layer_system.number_of_layers() - 1
        pwe_init_R = initial_field.plane_wave_expansion(layer_system, i_R)[0]
    
    pfe_sca = periodic_scattered_field_piecewise_expansion(initial_field, particle_list, layer_system, a1, a2, layer_numbers=[i_R])
    
    # layer response of all scattered fields
    pwe_R = copy.deepcopy(pfe_sca.expansion_list[-1])
    idx_a = np.argwhere(pwe_R.azimuthal_angles == pwe_init_R.azimuthal_angles)[0][0]
    idx_kpar = np.argwhere(pwe_R.k_parallel == pwe_init_R.k_parallel)[0][0]
    
    # add the initial field contribution
    pwe_R.coefficients[:, idx_kpar, idx_a] += pwe_init_R.coefficients.reshape(2)
    
    # if particles are located in the outer layer, add the direct scattered field contributions
    i_sca = layer_system.layer_number(particle_list[0].position[2]) # all particles have to be in the same layer
    if i_sca == i_R:
        for ip in range(len(particle_list)):
            if i_R == 0:
                pwe_R = pwe_R + pfe_sca.expansion_list[2 * ip + 1]
            else:
                pwe_R = pwe_R + pfe_sca.expansion_list[2 * ip]
 
    return pwe_R


# Field evaluation follows the procedure of Smuthi's "old (legacy)" electric field and magnetic field routines 
# basically identical to the field evaluation of any plane wave, but as a sum of discrete propagation angles (instead of an integral)
def electromagnetic_nearfield(pwe, x, y, z, chunksize=50, field_type='electric', vacuum_wavelength=None):
    """Evaluate electric or magnetic field.      
    Args:
        pwe (smuthi.field_expansion.PlaneWaveExpansion):    plane wave expansion emitted by a periodic particle
                                                            arrangement 
        x (numpy.ndarray):    x-coordinates of query points
        y (numpy.ndarray):    y-coordinates of query points
        z (numpy.ndarray):    z-coordinates of query points
        chunksize (int):      number of field points that are simultaneously evaluated when running in CPU mode
        field_type (str):     field type to be evaluated; either 'electric' or 'magnetic'
        vacuum_wavelength (float):  vacuum wavelength in length units (only necessary for magnetic fields).    
    Returns:
        Tuple of (E_x, E_y, E_z) or (H_x, H_y, H_z) numpy.ndarray objects with the Cartesian coordinates of
        complex electric field.
    """    
    emx = np.zeros(x.shape, dtype=complex)
    emy = np.zeros(x.shape, dtype=complex)
    emz = np.zeros(x.shape, dtype=complex)

    xr = x[pwe.valid(x, y, z)] - pwe.reference_point[0]
    yr = y[pwe.valid(x, y, z)] - pwe.reference_point[1]
    zr = z[pwe.valid(x, y, z)] - pwe.reference_point[2]
    
    if vacuum_wavelength:
        omega = flds.angular_frequency(vacuum_wavelength)
    
    if cu.use_gpu and xr.size and len(pwe.k_parallel) > 1:  # run calculations on gpu
            
        re_kp_d = gpuarray.to_gpu(pwe.k_parallel.real.astype(np.float32))
        im_kp_d = gpuarray.to_gpu(pwe.k_parallel.imag.astype(np.float32))
        
        re_kz_d = gpuarray.to_gpu(pwe.k_z().real.astype(np.float32))
        im_kz_d = gpuarray.to_gpu(pwe.k_z().imag.astype(np.float32))
        
        alpha_d = gpuarray.to_gpu(pwe.azimuthal_angles.astype(np.float32))
        
        xr_d = gpuarray.to_gpu(xr.astype(np.float32))
        yr_d = gpuarray.to_gpu(yr.astype(np.float32))
        zr_d = gpuarray.to_gpu(zr.astype(np.float32))
        
        re_g_te_d = gpuarray.to_gpu(pwe.coefficients[0, :, :].real.astype(np.float32))
        im_g_te_d = gpuarray.to_gpu(pwe.coefficients[0, :, :].imag.astype(np.float32))
        re_g_tm_d = gpuarray.to_gpu(pwe.coefficients[1, :, :].real.astype(np.float32))
        im_g_tm_d = gpuarray.to_gpu(pwe.coefficients[1, :, :].imag.astype(np.float32))
        
        re_em_x_d = gpuarray.to_gpu(np.zeros(xr.shape, dtype=np.float32))
        im_em_x_d = gpuarray.to_gpu(np.zeros(xr.shape, dtype=np.float32))
        re_em_y_d = gpuarray.to_gpu(np.zeros(xr.shape, dtype=np.float32))
        im_em_y_d = gpuarray.to_gpu(np.zeros(xr.shape, dtype=np.float32))
        re_em_z_d = gpuarray.to_gpu(np.zeros(xr.shape, dtype=np.float32))
        im_em_z_d = gpuarray.to_gpu(np.zeros(xr.shape, dtype=np.float32))
        
        
        if field_type == 'electric':
            kernel_source = cu_src.pwe_periodic_electric_field_evaluation_code%(xr.size, len(pwe.k_parallel), 
                                                                   len(pwe.azimuthal_angles), (1/pwe.k).real, 
                                                                   (1/pwe.k).imag)
            kernel_function = SourceModule(kernel_source).get_function("electric_field") 
            
            cuda_blocksize = 128
            cuda_gridsize = (xr.size + cuda_blocksize - 1) // cuda_blocksize
            
            kernel_function(re_kp_d, im_kp_d, re_kz_d, im_kz_d, alpha_d, xr_d, yr_d, zr_d, re_g_te_d, im_g_te_d,
                            re_g_tm_d, im_g_tm_d, re_em_x_d, im_em_x_d, re_em_y_d, im_em_y_d, re_em_z_d, im_em_z_d,
                            block=(cuda_blocksize,1,1), grid=(cuda_gridsize,1))
             
            emx[pwe.valid(x, y, z)] = re_em_x_d.get() + 1j * im_em_x_d.get()
            emy[pwe.valid(x, y, z)] = re_em_y_d.get() + 1j * im_em_y_d.get()
            emz[pwe.valid(x, y, z)] = re_em_z_d.get() + 1j * im_em_z_d.get()
            
        elif field_type == 'magnetic':
            kernel_source = cu_src.pwe_periodic_magnetic_field_evaluation_code%(xr.size, len(pwe.k_parallel), 
                                                                   len(pwe.azimuthal_angles), (pwe.k).real, 
                                                                   (pwe.k).imag)
            kernel_function = SourceModule(kernel_source).get_function("magnetic_field")
            
            cuda_blocksize = 128
            cuda_gridsize = (xr.size + cuda_blocksize - 1) // cuda_blocksize
            
            kernel_function(re_kp_d, im_kp_d, re_kz_d, im_kz_d, alpha_d, xr_d, yr_d, zr_d, re_g_te_d, im_g_te_d,
                            re_g_tm_d, im_g_tm_d, re_em_x_d, im_em_x_d, re_em_y_d, im_em_y_d, re_em_z_d, im_em_z_d,
                            block=(cuda_blocksize,1,1), grid=(cuda_gridsize,1))
             
            emx[pwe.valid(x, y, z)] = 1 / omega * (re_em_x_d.get() + 1j * im_em_x_d.get())
            emy[pwe.valid(x, y, z)] = 1 / omega * (re_em_y_d.get() + 1j * im_em_y_d.get())
            emz[pwe.valid(x, y, z)] = 1 / omega * (re_em_z_d.get() + 1j * im_em_z_d.get())
        else:
            assert False, 'Demanded field type must be electric or magnetic!'
            
    else: # run calculation of CPU
        kpgrid = pwe.k_parallel_grid()
        agrid = pwe.azimuthal_angle_grid()
        kx = kpgrid * np.cos(agrid)
        ky = kpgrid * np.sin(agrid)
        kz = pwe.k_z_grid()
    
        em_x_flat = np.zeros(xr.size, dtype=np.complex64)
        em_y_flat = np.zeros(xr.size, dtype=np.complex64)
        em_z_flat = np.zeros(xr.size, dtype=np.complex64)
        
        for i_chunk in range(int(np.ceil(xr.size / chunksize))):
            chunk_idcs = range(i_chunk * chunksize, min((i_chunk + 1) * chunksize, xr.size))
            xr_chunk = xr.flatten()[chunk_idcs]
            yr_chunk = yr.flatten()[chunk_idcs]
            zr_chunk = zr.flatten()[chunk_idcs]
    
            kr = np.zeros((len(xr_chunk), len(pwe.k_parallel), len(pwe.azimuthal_angles)), dtype=np.complex64)
            kr += np.tensordot(xr_chunk, kx, axes=0)
            kr += np.tensordot(yr_chunk, ky, axes=0)
            kr += np.tensordot(zr_chunk, kz, axes=0)
    
            eikr = np.exp(1j * kr)
            
            summand_x = np.zeros((len(xr_chunk), len(pwe.k_parallel), len(pwe.azimuthal_angles)),
                                   dtype=np.complex64)
            summand_y = np.zeros((len(yr_chunk), len(pwe.k_parallel), len(pwe.azimuthal_angles)),
                                   dtype=np.complex64)
            summand_z = np.zeros((len(zr_chunk), len(pwe.k_parallel), len(pwe.azimuthal_angles)),
                                   dtype=np.complex64)
            
            if field_type == 'electric':
                # pol=0
                summand_x += (-np.sin(agrid) * pwe.coefficients[0, :, :])[None, :, :] * eikr
                summand_y += (np.cos(agrid) * pwe.coefficients[0, :, :])[None, :, :] * eikr
                # pol=1
                summand_x += (np.cos(agrid) * kz / pwe.k * pwe.coefficients[1, :, :])[None, :, :] * eikr
                summand_y += (np.sin(agrid) * kz / pwe.k * pwe.coefficients[1, :, :])[None, :, :] * eikr
                summand_z += (-kpgrid / pwe.k * pwe.coefficients[1, :, :])[None, :, :] * eikr
                
                em_x_flat[chunk_idcs] = np.sum(np.sum(summand_x, axis=2), axis=1)
                em_y_flat[chunk_idcs] = np.sum(np.sum(summand_y, axis=2), axis=1)
                em_z_flat[chunk_idcs] = np.sum(np.sum(summand_z, axis=2), axis=1)        
            elif field_type == 'magnetic':
                 # pol=0
                summand_x += (- kz * np.cos(agrid) * pwe.coefficients[0, :, :])[None, :, :] * eikr
                summand_y += (- kz * np.sin(agrid) * pwe.coefficients[0, :, :])[None, :, :] * eikr
                summand_z += (kpgrid * pwe.coefficients[0, :, :])[None, :, :] * eikr
                # pol=1
                summand_x += (- np.sin(agrid) * pwe.k * pwe.coefficients[1, :, :])[None, :, :] * eikr
                summand_y += (np.cos(agrid) * pwe.k * pwe.coefficients[1, :, :])[None, :, :] * eikr 
                
                em_x_flat[chunk_idcs] = 1 / omega * np.sum(np.sum(summand_x, axis=2), axis=1)
                em_y_flat[chunk_idcs] = 1 / omega * np.sum(np.sum(summand_y, axis=2), axis=1)
                em_z_flat[chunk_idcs] = 1 / omega * np.sum(np.sum(summand_z, axis=2), axis=1)                
            else:
                assert False, 'Demanded field type must be electric or magnetic!'
                    
        emx[pwe.valid(x, y, z)] = em_x_flat.reshape(xr.shape)
        emy[pwe.valid(x, y, z)] = em_y_flat.reshape(xr.shape)
        emz[pwe.valid(x, y, z)] = em_z_flat.reshape(xr.shape)
    
    return emx, emy, emz


def conjugated_poynting_vector(E, H):
    """ 
    Args:
        E (tuple):  electric field tuple (E_x, E_y, E_z)
        H (tuple):  magnetic field tuple (H_x, H_y, H_z)
    Returns:
        Returns the complex conjugated Poynting vector S*.
    """
    ex_conj = np.conjugate(E[0])
    ey_conj = np.conjugate(E[1])
    ez_conj = np.conjugate(E[2])
    
    hx = H[0]
    hy = H[1]
    hz = H[2]
    
    sx = 1 / 2 * (ey_conj * hz - ez_conj * hy)
    sy = 1 / 2 * (ez_conj * hx - ex_conj * hz)
    sz = 1 / 2 * (ex_conj * hy - ey_conj * hx)
    
    return sx, sy, sz


# Note: The FarField.signal corresponds to the plane wave's power per interface area (Theobald 2021 dissertation, eq.(6.29))
# It differs from the smuthi.postprocessing.far_field.FarField object that is constructed from a PlaneWaveExpansion
# with a continuous angular distribution.
# Introduces the FarField.signal_type "normalized power".
def periodic_pwe_to_ff_conversion(plane_wave_expansion, initial_field, layer_system):
    """Compute the far field of a plane wave expansion object of discrete plane waves.
    Args:
        plane_wave_expansion (smuthi.fields.expansions.PlaneWaveExpansion):     
                                    plane wave expansion of the scattered field of a periodic particle
                                    arrangement that is to be convert into a far field object.
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        layer_system (smuthi.layer.LayerSystem):            stratified medium
    Returns:
        A smuthi.postprocessing.far_field.FarField object.
    """
    omega = flds.angular_frequency(initial_field.vacuum_wavelength)
    k = plane_wave_expansion.k
    kp = plane_wave_expansion.k_parallel
    if plane_wave_expansion.kind == 'upgoing':
        polar_angles = np.arcsin(kp[kp < k] / k)
    elif plane_wave_expansion.kind == 'downgoing':
        polar_angles = np.pi - np.arcsin((kp / k)[kp / k < 1])
    else:
        raise ValueError('PWE type not specified')
    if any(polar_angles.imag):
        raise ValueError('complex angles are not allowed')
    polar_angles = polar_angles[np.logical_not(np.isnan(polar_angles))]
    azimuthal_angles = plane_wave_expansion.azimuthal_angles
    
    # power per interface area (Theobald 2021 dissertation, eq.(6.29))
    normalized_power = (k / (2 * omega) * np.cos(polar_angles)[None, :, None] \
              * abs(plane_wave_expansion.coefficients[:, :len(polar_angles), :]) ** 2)
        
    srt_idcs = np.argsort(polar_angles)  # reversing order in case of downgoing
    ff = FarField(polar_angles=polar_angles[srt_idcs], azimuthal_angles=azimuthal_angles,
                  signal_type='normalized power')
    ff.signal = normalized_power[:, srt_idcs, :]
    return ff


def scattered_periodic_ff_power_per_area(far_field):
    """ Compute the scattered far field power per area of a periodic particle arangement.
    Args:
        far_field (smuthi.postprocessing.far_field.FarField):   Farfield object of a periodic, scattered SWE
    Returns:
        Scattered farfield power per area.
    """
    if not far_field.signal_type == 'normalized power':
        raise TypeError('Far field does not consist of discrete plane waves.')
    return abs(np.sum(far_field.signal))


# not explicitly connected to periodic particle arrangement and therefore could be a method of the class smuthi.initial_field.PlaneWave
def initial_plane_wave_power_per_area(initial_field, layer_system):
    """Compute the power carried by the initial plane wave into vertical direction per area.
    Args:
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        layer_system (smuthi.layer.LayerSystem):            stratified medium
    Returns:
        Initial fields power per area.
    """
    if 0 <= initial_field.polar_angle <= np.pi/2:
        n0 = layer_system.refractive_indices[0]
    elif np.pi / 2 <= initial_field.polar_angle < np.pi:
        n0 = layer_system.refractive_indices[-1]
    k0 = 2 * np.pi * n0 / initial_field.vacuum_wavelength
    omega = initial_field.angular_frequency()
    return  k0 / (2 * omega) * np.cos(initial_field.polar_angle) * initial_field.amplitude ** 2









"""Functions that assist the user in the choice of suitable numerical simulation parameters."""

import sys
import numpy as np
import matplotlib.pyplot as plt
import smuthi.fields as flds
import smuthi.postprocessing.far_field as ff
from smuthi.fields import angular_frequency
import smuthi.utility.logging as log


def evaluate(simulation, detector):
    """Run a simulation and evaluate the detector.
    Args:
        simulation (smuthi.simulation.Simulation):  simulation object
        detector (method or str):                   Specify a method that accepts a simulation as input and returns
                                                    a float. Otherwise, type "extinction cross section" to use the
                                                    extinction cross section as a detector.

    Returns:
        The detector value (float)
    """
    if detector == "extinction cross section":
        def detector(sim):
            ecs = ff.extinction_cross_section(initial_field=sim.initial_field,
                                              particle_list=sim.particle_list,
                                              layer_system=sim.layer_system)
            return ecs
    elif detector == "total scattering cross section":
        def detector(sim):
            scs = ff.total_scattering_cross_section(initial_field=sim.initial_field,
                                                    particle_list=sim.particle_list,
                                                    layer_system=sim.layer_system)
            return scs
    elif detector == "integrated scattered far field":
        def detector(sim):
            # _, _, scff = ff.total_far_field(initial_field=sim.initial_field,
            scff = ff.scattered_far_field(vacuum_wavelength=sim.initial_field.vacuum_wavelength,
                                          particle_list=sim.particle_list,
                                          layer_system=sim.layer_system)
            iscff = scff.integral()
            return iscff[0] + iscff[1]

    with log.LoggerMuted():
        simulation.run()
        return detector(simulation)

def update_lmax_mmax(simulation, l_max):
    """ Assign the same l_max and m_max = l_max to all particles in simulation """
    for particle in simulation.particle_list:
        particle.l_max = l_max
        particle.m_max = l_max

def converge_l_max(simulation,
                   detector="extinction cross section",
                   tolerance=1e-3,
                   tolerance_steps=2,
                   max_iter=100,
                   start_from_1=True,
                   ax=None):
    """Find suitable multipole cutoff degree `l_max` for a given particle and simulation. The routine starts with the
    current `l_max` of the particle. The value of `l_max` is successively incremented in a loop until the resulting
    relative change in the detector value is smaller than the specified tolerance. The method updates the input
    particle object with the `l_max` value for which convergence has been achieved.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object containing the particle
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence.
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value.
        tolerance (float):                          Relative tolerance for the detector value change.
        tolerance_steps (int):                      Number of consecutive steps at which the tolerance must be met
                                                    during multipole truncation convergence. Default: 2
        max_iter (int):                             Break convergence loop after that number of iterations, even if
                                                    no convergence has been achieved.
        start_from_1 (logical):                     If true (default), start from `l_max=1`. Otherwise, start from the
                                                    current particle `l_max`.
        ax (np.array of AxesSubplot):               Array of AxesSubplots where to live-plot convergence output

    Returns:
        A 3-tuple containing 
          - detector value of converged or break-off parameter settings.
          - series of lmax values
          - the detector values for the given lmax values
      """

    print("")
    print("------------------------")
    log.write_blue("Searching suitable l_max")

    if start_from_1:
        l_max = 1
        update_lmax_mmax(simulation, l_max)
    else:
        l_max = simulation.particle_list[0].l_max # assuming that all particles have the same l_max

    print("Start value: l_max=%i" % l_max)

    current_value = evaluate(simulation, detector)

    x = np.array([l_max])
    y = np.array([current_value])
    r = np.array([])
    if ax is not None:
        line1, = ax[0].plot(x, np.real(y), '.-')
        line2, = ax[1].plot(x[1:], r, '.-')

    for _ in range(max_iter):
        l_max += 1  # l_max increment
        update_lmax_mmax(simulation, l_max)

        print("---------------------------------------")
        print("Try l_max = %i and m_max=%i" % (l_max, l_max))

        new_value = evaluate(simulation, detector)
        rel_diff = abs(new_value - current_value) / abs(current_value)
        print("Old detector value:", current_value)
        print("New detector value:", new_value)
        print("Relative difference:", rel_diff)
        print("Allowed tolerance:  ", tolerance)

        neff_max = simulation.neff_max
        x = np.append(x, l_max)
        y = np.append(y, new_value)
        r = np.append(r, rel_diff)
        if ax is not None:
            line1.set_data(x, np.real(y))
            if neff_max is not None:
                line1.set_label('$n_{eff}^{max} = %g$'%neff_max.real)
            else:
                line1.set_label('$l_{max}$')
            line2.set_data(x[1:], r)
            for axi in ax:
                axi.relim()
                axi.autoscale_view()
            plt.draw()
            plt.pause(0.001)
            ax[0].legend()

        if np.all(r[-tolerance_steps:] < tolerance) and len(r) >= tolerance_steps: # then, discard l_max increment
            old_l_max = l_max - tolerance_steps
            update_lmax_mmax(simulation, old_l_max)
            log.write_green("Relative difference smaller than tolerance. Keep l_max = %i" % old_l_max)
            return y[-tolerance_steps-1], x, y
        else:
            current_value = new_value

    log.write_red("No convergence achieved. Keep l_max = %i"%l_max)
    return current_value, x, y


def converge_m_max(simulation,
                   detector="extinction cross section",
                   tolerance=1e-3,
                   target_value=None,
                   ax=None):
    """Find suitable multipole cutoff order `m_max` for a given particle and simulation. The routine starts with the
    current `l_max` of the particle, i.e. with `m_max=l_max`. The value of `m_max` is successively decremented in a loop
    until the resulting relative change in the detector value is larger than the specified tolerance. The method updates
    the input particle object with the so determined `m_max`.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object containing the particle
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence.
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value.
        tolerance (float):                          Relative tolerance for the detector value change.
        max_iter (int):                             Break convergence loop after that number of iterations, even if
                                                    no convergence has been achieved.
        target_value (float):                       If available (typically from preceding neff selection procedure),
                                                    use as target detector value
        ax (np.array of AxesSubplot):               Array of AxesSubplots where to live-plot convergence output

    Returns:
        Detector value of converged or break-off parameter settings.
    """

    def update_mmax(simulation, m_max):
        """ Assign the same m_max to all particles in simulation """
        for particle in simulation.particle_list:
            particle.m_max = m_max

    print("")
    print("------------------------")
    log.write_blue("Searching suitable m_max")
    m_max = simulation.particle_list[0].m_max # assuming that all particles have the same l_max
    print("Start value: m_max=%i"%m_max)

    current_value = evaluate(simulation, detector) if target_value is None else target_value

    x = np.array([])
    y = np.array([])
    r = np.array([])
    if ax is not None:
        ax[0].axhline(current_value, color='black', linestyle='dashed',
                      label='target value')
        line1, = ax[0].plot(x, y, '.-')
        line2, = ax[1].plot(x, r, '.-')

    for m_max in range(m_max, -1, -1):
        old_m_max = simulation.particle_list[0].m_max
        update_mmax(simulation, m_max)
        print("---------------------------------------")
        print("Try m_max=%i"%m_max)
        new_value = evaluate(simulation, detector)
        rel_diff = abs(new_value - current_value) / abs(current_value)
        print("Old detector value:", current_value)
        print("New detector value:", new_value)
        print("Relative difference:", rel_diff)
        print("Allowed tolerance:  ", tolerance)

        x = np.append(x, m_max)
        y = np.append(y, new_value.real)
        r = np.append(r, rel_diff)
        if ax is not None:
            line1.set_data(x, y)
            line1.set_label('$m_{max}$ for $l_{max} = %g$'%simulation.particle_list[0].l_max)
            line2.set_data(x, r)
            for axi in ax:
                axi.relim()
                axi.autoscale_view()
            plt.draw()
            plt.pause(0.001)
            ax[0].legend()

        if np.any(r > tolerance):  # in this case: discard m_max decrement
            update_mmax(simulation, old_m_max)
            log.write_green("Relative difference larger than tolerance. Keep m_max = %g"%old_m_max)
            if ax is not None:
                titlestr = "relative diff > {:g}, keep $m_{{max}} = {:g}$".format(tolerance, old_m_max)
                ax[1].title.set_text(titlestr)
                ax[1].title.set_color('g')
                plt.draw()
            return current_value
        else:
            titlestr = "no convergence achieved, keep $m_{{max}} = {:g}$".format(m_max)
            current_value = new_value if target_value is None else target_value

    log.write_red("No convergence achieved. Keep m_max = %g" % m_max)
    if ax is not None:
        ax[1].title.set_text(titlestr)
        ax[1].title.set_color('r')
        plt.draw()


def converge_multipole_cutoff(simulation,
                              detector="extinction cross section",
                              tolerance=1e-3,
                              tolerance_steps=2,
                              max_iter=100,
                              current_value=None,
                              l_max_list=None,
                              detector_value_list=None,
                              converge_m=True,
                              ax=None):
    """Find suitable multipole cutoff degree `l_max` and order `m_max` for all particles in a given simulation object.
    The method updates the input simulation object with the so determined multipole truncation values.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value
        tolerance (float):                          Relative tolerance for the detector value change
        tolerance_steps (int):                      Number of consecutive steps at which the tolerance must be met
                                                    during multipole truncation convergence. Default: 2
        max_iter (int):                             Break convergence loops after that number of iterations, even if
                                                    no convergence has been achieved
        current_value (float):                      If specified, skip l_max run and use this value for the
                                                    resulting detector value.
                                                    Otherwise, start with l_max run.
        l_max_list (list):                          If current_value was specified, the l_max run is skipped.
                                                    Then, this list is returned as the second item in the returned tuple.
        detector_value_list (list):                 If current_value was specified, the l_max run is skipped.
                                                    Then, this list is returned as the third item in the returned tuple.
        converge_m (logical):                       If false, only converge `l_max`, but keep `m_max=l_max`. Default
                                                    is true
        ax (np.array of AxesSubplot):               Array of AxesSubplots where to live-plot convergence output

    Returns:
        A 3-tuple containing
          - detector value of converged or break-off parameter settings.
          - series of lmax values
          - the detector values for the given lmax values
    """
    print("")
    print("-----------------------------------")
    log.write_blue("Searching suitable multipole cutoff")

    with log.LoggerIndented():
        if current_value is None:
            current_value, lmx, values = converge_l_max(simulation=simulation,
                                                        detector=detector,
                                                        tolerance=tolerance,
                                                        tolerance_steps=tolerance_steps,
                                                        max_iter=max_iter,
                                                        start_from_1=True, # start_from_1
                                                        ax=ax)
        else:
            log.write_green("Keep l_max = %i"%simulation.particle_list[0].l_max)
            lmx = l_max_list
            values = detector_value_list

        if converge_m:
            current_value = converge_m_max(simulation=simulation,
                                           detector=detector,
                                           tolerance=tolerance,
                                           target_value=current_value, # passing current_value as the target_value
                                           ax=ax)

    return current_value, lmx, values


def update_contour(simulation, neff_imag=5e-3, neff_max=None, neff_max_offset=0.5, neff_resolution=2e-3):
    """Update the default `k_parallel` arrays in smuthi.fields with a newly constructed Sommerfeld integral
    contour, and set the simulation object to use the default contour for particle coupling.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        neff_imag (float):                          Extent of the contour into the negative imaginary direction
                                                    (in terms of effective refractive index, n_eff=kappa/omega).
        neff_max (float):                           Truncation value of contour (in terms of effective refractive
                                                    index).
        neff_max_offset (float):                    If no value for `neff_max` is specified, use the last estimated
                                                    singularity location plus this value (in terms of effective
                                                    refractive index).
        neff_resolution (float):                    Discretization of the contour (in terms of effective refractive
                                                    index).
    """
    simulation.neff_imag = neff_imag
    simulation.neff_max = neff_max
    simulation.neff_max_offset = neff_max_offset
    simulation.neff_resolution = neff_resolution
    simulation.set_default_initial_field_contour()
    simulation.set_default_Sommerfeld_contour()

    simulation.k_parallel = "default"


def converge_neff_max(simulation,
                      detector="extinction cross section",
                      tolerance=1e-3,
                      tolerance_factor=0.1,
                      tolerance_steps=2,
                      max_iter=30,
                      neff_imag=1e-2,
                      neff_resolution=2e-3,
                      neff_max_increment=0.5,
                      neff_max_offset=0,
                      converge_lm=True,
                      ax=None):
    """Find a suitable truncation value for the multiple scattering Sommerfeld integral contour and update the
    simulation object accordingly.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence.
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value.
        tolerance (float):                          Relative tolerance for the detector value change.
        tolerance_factor (float):                   During neff selection, a smaller tolerance should be allowed to
                                                    avoid fluctuations of the order of ~tolerance which would
                                                    compromise convergence. Default: 0.1
        tolerance_steps (int):                      Number of consecutive steps at which the tolerance must be met
                                                    during multipole truncation convergence. Default: 2
        max_iter (int):                             Break convergence loops after that number of iterations, even if
                                                    no convergence has been achieved.
        neff_imag (float):                          Extent of the contour into the negative imaginary direction
                                                    (in terms of effective refractive index, n_eff=kappa/omega).
        neff_resolution (float):                    Discretization of the contour (in terms of eff. refractive
                                                    index).
        neff_max_increment (float):                 Increment the neff_max parameter with that step size
        neff_max_offset (float):                    Start neff_max selection from the last estimated singularity
                                                    location plus this value (in terms of effective refractive index)
        converge_lm (logical):                      If set to true, update multipole truncation during each step
                                                    (this takes longer time, but is necessary for critical use cases
                                                    like flat particles on a substrate)
        ax (np.array of AxesSubplot):               Array of AxesSubplots where to live-plot convergence output

    Returns:
        Detector value for converged settings.
    """

    print("")
    print("---------------------------")
    log.write_blue("Searching suitable neff_max")

    update_contour(simulation=simulation, neff_imag=neff_imag, neff_max_offset=neff_max_offset, neff_resolution=neff_resolution)

    simulation.k_parallel = flds.reasonable_Sommerfeld_kpar_contour(
        vacuum_wavelength=simulation.initial_field.vacuum_wavelength,
        layer_refractive_indices=simulation.layer_system.refractive_indices,
        neff_imag=neff_imag,
        neff_max_offset=neff_max_offset,
        neff_resolution=neff_resolution)

    neff_max = simulation.k_parallel[-1] / angular_frequency(simulation.initial_field.vacuum_wavelength)
    simulation.neff_max = neff_max
    print("Starting value: neff_max=%f"%neff_max.real)

    if converge_lm:
        with log.LoggerIndented():
            current_value, lmx_list, values = converge_multipole_cutoff(simulation=simulation,
                                                                        detector=detector,
                                                                        tolerance=tolerance*tolerance_factor,
                                                                        tolerance_steps=tolerance_steps,
                                                                        max_iter=max_iter,
                                                                        converge_m=False,
                                                                        ax=ax)
        lmx_lists = [lmx_list]
        lmx_values_lists = [values]

    else:
        current_value = evaluate(simulation, detector)

    x = np.array([neff_max.real])
    y = np.array([current_value.real])
    r = np.array([])
    if ax is not None:
        line1, = ax[-2].plot(x, y, 'k.-')
        line2, = ax[-1].plot(x[1:], r, 'k.-')
        ax[-1].set_xlabel('$n_{eff}^{max}$')
        ax[-2].set_title('$n_{eff}^{max}$ selection')

    for _ in range(max_iter):
        old_neff_max = neff_max
        neff_max = neff_max + neff_max_increment
        update_contour(simulation=simulation, neff_imag=neff_imag, neff_max=neff_max, neff_resolution=neff_resolution)

        print("---------------------------------------")
        print("Try neff_max = %f"%neff_max.real)

        if converge_lm:
            with log.LoggerIndented():
                new_value, lmx_list, values = converge_multipole_cutoff(simulation=simulation,
                                                                        detector=detector,
                                                                        tolerance=tolerance*tolerance_factor,
                                                                        tolerance_steps=tolerance_steps,
                                                                        max_iter=max_iter,
                                                                        current_value=None,
                                                                        converge_m=False,
                                                                        ax=ax)
            lmx_lists.append(lmx_list)
            lmx_values_lists.append(values)
        else:
            new_value = evaluate(simulation, detector)


        rel_diff = abs(new_value - current_value) / abs(current_value)
        print("Old detector value:", current_value)
        print("New detector value:", new_value)
        print("Relative difference:", rel_diff)
        print("Allowed tolerance:  ", tolerance)

        x = np.append(x, neff_max.real)
        y = np.append(y, new_value.real)
        r = np.append(r, rel_diff)
        if ax is not None:
            line1.set_data(x, y)
            line1.set_label('$n_{eff}^{max}$')
            line2.set_data(x[1:], r)
            for axi in ax:
                axi.relim()
                axi.autoscale_view()
            plt.draw()
            plt.pause(0.001)
            for axi in ax[::-2]:
                axi.legend()

        if rel_diff < tolerance:  # then, discard neff_max increment
            neff_max = old_neff_max
            update_contour(simulation=simulation, neff_imag=neff_imag, neff_max=neff_max, neff_resolution=neff_resolution)
            log.write_green(
                "Relative difference smaller than tolerance. Keep neff_max = %g" % neff_max.real)
            if converge_lm:
                # until now, we have looked for convergence with regard to reduced tolerance (by tolerance factor)
                # now, we check for the smallest l_max value that satisfies the actual tolerance criterium
                for idx, lmx in enumerate(lmx_lists[-2]):
                    update_lmax_mmax(simulation, lmx)
                    current_value = lmx_values_lists[-2][idx]
                    all_fine = True
                    print("check ", lmx_lists[-2][idx])
                    for incr in range(1, tolerance_steps + 1):
                        rel_diff = abs(current_value - lmx_values_lists[-2][idx + incr]) / abs(lmx_values_lists[-2][idx + incr])
                        if rel_diff > tolerance:
                            all_fine = False
                    if all_fine:
                        break

            if ax is not None:
                titlestr = "relative diff < {:g}, keep $n_{{eff}}^{{max}} = {:g}$".format(tolerance, neff_max.real)
                ax[-1].title.set_text(titlestr)
                ax[-1].title.set_color('g')
                plt.draw()

            return current_value  # most accurate result for old_neff_max
        else:
            titlestr = "no convergence achieved, keep $n_{{eff}}^{{max}} = {:g}$".format(neff_max.real)
            current_value = new_value

    log.write_red("No convergence achieved. Keep neff_max = %g" % neff_max.real)
    if ax is not None:
        ax[-1].title.set_text(titlestr)
        ax[-1].title.set_color('r')
        plt.draw()

    return current_value


def converge_neff_resolution(simulation,
                       detector="extinction cross section",
                       tolerance=1e-3,
                       max_iter=30,
                       neff_imag=1e-2,
                       neff_max=None,
                       neff_resolution=1e-2,
                       ax=None):
    """Find a suitable discretization step size for the multiple scattering Sommerfeld integral contour and update
    the simulation object accordingly.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence.
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value.
        tolerance (float):                          Relative tolerance for the detector value change.
        max_iter (int):                             Break convergence loops after that number of iterations, even if
                                                    no convergence has been achieved.
        neff_imag (float):                          Extent of the contour into the negative imaginary direction
                                                    (in terms of effective refractive index, n_eff=kappa/omega).
        neff_max (float):                           Truncation value of contour (in terms of effective refractive
                                                    index).
        neff_resolution (float):                    Discretization of the contour (in terms of eff. refractive
                                                    index) - start value for iteration
        ax (np.array of AxesSubplot):               Array of AxesSubplots where to live-plot convergence output

    Returns:
        Detector value for converged settings.
    """
    print("")
    print("-----------------------")
    log.write_blue("Find suitable neff_resolution")

    update_contour(simulation=simulation, neff_imag=neff_imag, neff_max=neff_max, neff_resolution=neff_resolution)
    print("Starting value: neff_resolution=%f" % neff_resolution)
    current_value = evaluate(simulation, detector)

    if ax is not None:
        x = np.array([])
        y = np.array([])
        r = np.array([])
        init_x = neff_resolution
        init_y = current_value.real
        line1, = ax[0].plot(np.insert(x, 0, init_x), np.insert(y.real, 0, init_y),'.-')
        line2, = ax[1].plot(x,r,'.-')

    for _ in range(max_iter):
        old_neff_resolution = neff_resolution
        neff_resolution = neff_resolution / 2.0
        update_contour(simulation=simulation, neff_imag=neff_imag, neff_max=neff_max, neff_resolution=neff_resolution)

        print("---------------------------------------")
        print("Try neff_resolution = %f" % neff_resolution)

        new_value = evaluate(simulation, detector)

        rel_diff = abs(new_value - current_value) / abs(current_value)
        print("Old detector value:", current_value)
        print("New detector value:", new_value)
        print("Relative difference:", rel_diff)
        print("Allowed tolerance:  ", tolerance)

        if ax is not None:
            x = np.append(x, neff_resolution)
            y = np.append(y, new_value)
            r = np.append(r, rel_diff)
            line1.set_data(np.insert(x, 0, init_x), np.insert(y.real, 0, init_y.real))
            line1.set_label('$n_{{eff}}^{{max}} = {:g}$, $l_{{max}} = {}$, $m_{{max}} = {}$'.\
                            format(simulation.neff_max.real,
                                   simulation.particle_list[0].l_max,
                                   simulation.particle_list[0].m_max))
            line2.set_data(x, r)
            [( ax.relim(), ax.autoscale_view()) for ax in ax]
            plt.draw()
            plt.pause(0.001)
            ax[0].legend()

        if rel_diff < tolerance:  # in this case: discard halfed neff_resolution
            neff_resolution = old_neff_resolution
            update_contour(simulation=simulation, neff_imag=neff_imag, neff_max=neff_max, neff_resolution=neff_resolution)
            log.write_green("Relative difference smaller than tolerance. Keep neff_resolution = %g" % neff_resolution)
            if ax is not None:
                titlestr = "relative diff < {:g}, keep $\delta n_{{eff}} = {:g}$".format(tolerance, neff_resolution)
                ax[1].title.set_text(titlestr)
                ax[1].title.set_color('g')
                plt.draw()
            return current_value
        else:
            titlestr = "no convergence achieved, keep $\delta n_{{eff}} = {:g}$".format(neff_max.real)
            current_value = new_value

    log.write_red("No convergence achieved. Keep neff_resolution = %g" % neff_resolution)
    if ax is not None:
        ax[1].title.set_text(titlestr)
        ax[1].title.set_color('r')
        plt.draw()

    return None


def converge_angular_resolution(simulation,
                                detector="extinction cross section",
                                tolerance=1e-3,
                                max_iter=30,
                                ax=None):
    """Find a suitable discretization step size for the default angular arrays used for plane wave expansions.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence.
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value.
        tolerance (float):                          Relative tolerance for the detector value change.
        max_iter (int):                             Break convergence loops after that number of iterations, even if
                                                    no convergence has been achieved.
        ax (np.array of AxesSubplot):               Array of AxesSubplots where to live-plot convergence output

    Returns:
        Detector value for converged settings.
    """

    print("")
    print("-----------------------")
    log.write_blue("Find suitable angular_resolution")

    angular_resolution = simulation.angular_resolution # NOTE: should we have a starting-value flag?
    print("Starting value: angular_resolution=%f rad" % angular_resolution)
    current_value = evaluate(simulation, detector)

    if ax is not None:
        x = np.array([])
        y = np.array([])
        r = np.array([])
        init_x = angular_resolution
        init_y = current_value.real
        line1, = ax[0].plot(np.insert(x, 0, init_x), np.insert(y.real, 0, init_y),'.-')
        line2, = ax[1].plot(x,r,'.-')

    for _ in range(max_iter):
        old_angular_resolution = angular_resolution
        angular_resolution = angular_resolution / 2.0
        simulation.angular_resolution = angular_resolution
        simulation.set_default_angles()

        print("---------------------------------------")
        print("Try angular_resolution = %f rad" % angular_resolution)

        new_value = evaluate(simulation, detector)

        rel_diff = abs(new_value - current_value) / abs(current_value)
        print("Old detector value:", current_value)
        print("New detector value:", new_value)
        print("Relative difference:", rel_diff)
        print("Allowed tolerance:  ", tolerance)

        if ax is not None:
            x = np.append(x, angular_resolution)
            y = np.append(y, new_value)
            r = np.append(r, rel_diff)
            line1.set_data(np.insert(x, 0, init_x), np.insert(y.real, 0, init_y.real))
            line1.set_label('$n_{{eff}}^{{max}} = {:g}$, $l_{{max}} = {}$, $m_{{max}} = {}$, $\delta n_{{eff}} = {}$'.\
                            format(simulation.neff_max.real,
                                   simulation.particle_list[0].l_max,
                                   simulation.particle_list[0].m_max,
                                   simulation.neff_resolution))
            line2.set_data(x, r)
            [( ax.relim(), ax.autoscale_view()) for ax in ax]
            plt.draw()
            plt.pause(0.001)
            ax[0].legend()

        if rel_diff < tolerance:  # in this case: discard halfed neff_resolution
            simulation.angular_resolution = old_angular_resolution
            simulation.set_default_angles()
            log.write_green("Relative difference smaller than tolerance. Keep angular_resolution = %g deg" % old_angular_resolution)
            if ax is not None:
                titlestr = "relative diff < {:g}, keep $\delta \\theta = {:g}$ deg".format(tolerance, old_angular_resolution)
                ax[1].title.set_text(titlestr)
                ax[1].title.set_color('g')
                plt.draw()
            return current_value
        else:
            titlestr = "no convergence achieved, keep $\delta \\theta = {:g}$ deg".format(angular_resolution)
            current_value = new_value

    log.write_red("No convergence achieved. Keep angular_resolution = %g deg" % angular_resolution)
    if ax is not None:
        ax[1].title.set_text(titlestr)
        ax[1].title.set_color('r')
        plt.draw()

    return None


def select_numerical_parameters(simulation,
                                detector="extinction cross section",
                                tolerance=1e-3,
                                tolerance_factor=0.1,
                                tolerance_steps=2,
                                max_iter=30,
                                neff_imag=1e-2,
                                neff_resolution=1e-2,
                                select_neff_max=True,
                                neff_max_increment=0.5,
                                neff_max_offset=0,
                                neff_max=None,
                                select_neff_resolution=True,
                                select_angular_resolution=False,
                                select_multipole_cutoff=True,
                                relative_convergence=True,
                                show_plot=True):
    """Trigger automatic selection routines for various numerical parameters.

    Args:
        simulation (smuthi.simulation.Simulation):  Simulation object from which parameters are read and into which
                                                    results are stored.
        detector (function or string):              Function that accepts a simulation object and returns a detector
                                                    value the change of which is used to define convergence.
                                                    Alternatively, use "extinction cross section" (default) to have
                                                    the extinction cross section as the detector value.
        tolerance (float):                          Relative tolerance for the detector value change.
        tolerance_factor (float):                   During neff selection, a smaller tolerance should be allowed to
                                                    avoid fluctuations of the order of ~tolerance which would
                                                    compromise convergence. Default: 0.1
        tolerance_steps (int):                      Number of consecutive steps at which the tolerance must be met
                                                    during multipole truncation convergence. Default: 2
        max_iter (int):                             Break convergence loops after that number of iterations, even if
                                                    no convergence has been achieved.
        neff_imag (float):                          Extent of the contour into the negative imaginary direction
                                                    (in terms of effective refractive index, n_eff=kappa/omega)
        neff_resolution (float):                    Discretization of the contour (in terms of eff. refractive
                                                    index) - if `select_neff_resolution` is true, this value will be
                                                    eventually overwritten. However, it is required in any case.
                                                    Default: 1e-2
        select_neff_max (logical):                  If set to true (default), the Sommerfeld integral truncation
                                                    parameter `neff_max` is determined automatically with the help
                                                    of a Cauchy convergence criterion.
        neff_max_increment (float):                 Only needed if `select_neff_max` is true.
                                                    Step size with which `neff_max` is incremented.
        neff_max_offset (float):                    Only needed if `select_neff_max` is true.
                                                    Start n_eff selection from the last estimated singularity
                                                    location plus this value (in terms of effective refractive index)
        neff_max (float):                           Only needed if `select_neff_max` is false.
                                                    Truncation value of contour (in terms of effective refractive
                                                    index).
        select_neff_resolution (logical):           If set to true (default), the Sommerfeld integral discretization
                                                    parameter `neff_resolution` is determined automatically with the
                                                    help of a Cauchy convergence criterion.
        select_angular_resolution (logical):        If set to true, the angular resolution step for the default polar
                                                    and azimuthal angles is determined automatically according to a
                                                    Cauchy convergenge criterion.
        select_multipole_cutoff (logical):          If set to true (default), the multipole expansion cutoff
                                                    parameters `l_max` and `m_max` are determined automatically with
                                                    the help of a Cauchy convergence criterion.
        relative_convergence (logical):             If set to true (default), the `neff_max` convergence and the
                                                    `l_max` and `m_max` convergence routine are performed in the
                                                    spirit of relative convergence, i.e., the multipole expansion
                                                    convergence is checked again for each value of the Sommerfeld
                                                    integral truncation. This takes more time, but is required at
                                                    least in the case of flat particles near interfaces.
    """

    print("")
    print("----------------------------------------")
    print("Starting automatic paramateter selection")
    print("----------------------------------------")

    if tolerance_factor <= 0 or tolerance_factor > 1:
        raise ValueError('please provide 0 < tolerance_factor <= 1')

    if select_neff_max:
        if relative_convergence:
            if show_plot:
                plt.ion()
                _, ax_array = _init_fig('$l_{max}$', detector,
                                        'multipole cutoff',
                                        tolerance, tolerance_factor * tolerance,
                                        2)
            else:
                ax_array = None

            target_value = converge_neff_max(simulation=simulation,
                                             detector=detector,
                                             tolerance=tolerance,
                                             tolerance_factor=tolerance_factor,
                                             tolerance_steps=tolerance_steps,
                                             max_iter=max_iter,
                                             neff_imag=neff_imag,
                                             neff_resolution=neff_resolution,
                                             neff_max_increment=neff_max_increment,
                                             neff_max_offset=neff_max_offset,
                                             converge_lm=relative_convergence,
                                             ax=ax_array)
            neff_max = simulation.neff_max

            if select_multipole_cutoff:
                if show_plot:
                    plt.ion()
                    _, ax_array = _init_fig('multipole order', detector,
                                            'multipole cutoff selection',
                                            tolerance)
                else:
                    ax_array = None
                converge_multipole_cutoff(simulation=simulation,
                                          detector=detector,
                                          tolerance=tolerance,
                                          tolerance_steps=tolerance_steps,
                                          max_iter=max_iter,
                                          current_value=target_value,
                                          ax=ax_array)

        else:  # select neff_max but no relative convergence
            if select_multipole_cutoff:
                if show_plot:
                    plt.ion()
                    _, ax_array = _init_fig('multipole order', detector,
                                            'multipole cutoff selection',
                                            tolerance)
                else:
                    ax_array = None
                converge_multipole_cutoff(simulation=simulation,
                                          detector=detector,
                                          tolerance=tolerance,
                                          tolerance_steps=tolerance_steps,
                                          max_iter=max_iter,
                                          current_value=None,
                                          ax=ax_array)

            if show_plot:
                plt.ion()
                _, ax_array = _init_fig('$l_{max}$', detector,
                                        'multipole cutoff',
                                        tolerance, tolerance_factor * tolerance,
                                        2)
            else:
                ax_array = None

            target_value = converge_neff_max(simulation=simulation,
                                             detector=detector,
                                             tolerance=tolerance,
                                             tolerance_factor=tolerance_factor,
                                             tolerance_steps=tolerance_steps,
                                             max_iter=max_iter,
                                             neff_imag=neff_imag,
                                             neff_resolution=neff_resolution,
                                             neff_max_increment=neff_max_increment,
                                             neff_max_offset=neff_max_offset,
                                             converge_lm=relative_convergence,
                                             ax=ax_array)
            neff_max = simulation.neff_max

    else:  # no neff_max selection
        ax_array = None
        if select_multipole_cutoff:
            if show_plot:
                plt.ion()
                _, ax_array = _init_fig('multipole order', detector,
                                        'multipole cutoff selection',
                                        tolerance)
            else:
                ax_array = None
            converge_multipole_cutoff(simulation=simulation,
                                      detector=detector,
                                      tolerance=tolerance,
                                      tolerance_steps=tolerance_steps,
                                      max_iter=max_iter,
                                      current_value=None,
                                      ax=ax_array)


    if select_neff_resolution:
        if show_plot:
            plt.ion()
            _, ax_array = _init_fig('$\delta n_{eff}$', detector, '$\delta n_{eff}$ selection', tolerance)
        else:
            ax_array = None
        converge_neff_resolution(simulation=simulation,
                                 detector=detector,
                                 tolerance=tolerance,
                                 max_iter=max_iter,
                                 neff_imag=neff_imag,
                                 neff_max=neff_max,
                                 ax=ax_array)

    if select_angular_resolution:
        if show_plot:
            plt.ion()
            _, ax_array = _init_fig('$\delta \\theta$ [deg]', detector, 'angular resolution selection', tolerance)
        else:
            ax_array = None
        converge_angular_resolution(simulation=simulation,
                                    detector=detector,
                                    tolerance=tolerance,
                                    max_iter=max_iter,
                                    ax=ax_array)

    if show_plot:
        plt.ioff()


def _init_fig(xlabel='x', ylabel='y', title=None, tol=None, allowedtol=None, cols=1):
    """
    Inner utility function returning figure and axes handles initialized
    following a common template for all parameter selection runs.

    Args:
        xlabel (string):    column-wise common x axis label
        ylabel (string):    y axis label for the detector quantity
        title  (string):    title string for the (lefmost, if cols=2) detector panel
        tol (float):        user-set tolerance level to be drawn in relative difference panel
        allowedtol (float): when specified, also draw the actual tolerance level for the current run
        cols (int):         number of subplot columns. select_neff_max() accepts 2
    """
    fig, ax_array = plt.subplots(2, cols, sharex='col', figsize=(cols*6.4,8), gridspec_kw={'height_ratios': [3, 1]})
    ax_array = ax_array.flatten(order='F')
    ax_array[1].set_xlabel(xlabel)
    ax_array[0].set_ylabel(ylabel)
    ax_array[1].set_ylabel('relative difference')
    ax_array[0].set_title(title)
    [ax.axhline(tol, color='grey', linestyle='dashed', label='tolerance') for ax in ax_array[::-2]]
    if allowedtol is not None:
        ax_array[1].axhline(allowedtol, color='grey', linestyle='dotted', label='allowed tolerance')
    [( ax.set_yscale('log'), ax.grid(), ax.legend() ) for ax in ax_array[::-2]]

    return fig, ax_array


import numpy as np
from numba import complex64, complex128, float64, prange, jit, int32, int64, types


@jit(['complex64[:](complex64[:, :, :], float64[:], int32)',
        'complex128[:](complex128[:, :, :], float64[:], int32)'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=False)
def __numba_trapz_3dim_array_linear_part(y, x, i):
    one_dim_result = np.zeros((y.shape[1]), dtype=type(y[0, 0, 0]))
    for j in range(y.shape[1]):
        out = .0 + .0j
        for k in range(len(y[i, j]) - 1):
            out += (x[k+1]-x[k]) * (y[i, j, k+1] + y[i, j, k])/2.0
        one_dim_result[j] = out

    return one_dim_result


@jit(['void(complex64[:, :, :], complex64[:, :, :], complex64[:, :, :], complex64[:, :, :], \
            complex64[:, :, :], complex64[:, :, :], complex64[:, :, :], int32, int32)',
        'void(complex128[:, :, :], complex128[:, :, :], complex128[:, :, :], complex128[:, :, :], \
            complex128[:, :, :], complex128[:, :, :], complex128[:, :, :], int32, int32)'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=False)
def __evaluate_r_times_eikr_linear_part(foo_x, foo_y, foo_z, exp_feed, result_x, result_y, result_z, j, i_shape):
    foo_x_slice = foo_x[0, j]
    foo_y_slice = foo_y[0, j]
    foo_z_slice = foo_z[0, j]
    for i in range(i_shape):
        exp_j = (np.cos(exp_feed[i, j].real) + 1.0j*np.sin(exp_feed[i, j].real)) * \
                                                    np.exp(-exp_feed[i, j].imag)
        result_x[i, j] = foo_x_slice * exp_j
        result_y[i, j] = foo_y_slice * exp_j
        result_z[i, j] = foo_z_slice * exp_j

@jit(['complex64[:, :](complex64[:, :, :], float64[:])',
        'complex128[:, :](complex128[:, :, :], float64[:])'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=False)
def numba_trapz_3dim_array(y, x):
    """
    This function can replace snippet

    'foo = np.trapz(y, x)'

    by

    'foo = numba_trapz_3dim_array(y, x)'
    """
    result = np.zeros((y.shape[0], y.shape[1]), dtype=type(y[0, 0, 0]))
    for i in range(y.shape[0]):
        result[i] =__numba_trapz_3dim_array_linear_part(y, x, i)

    return result


@jit(['complex64[:, :](complex64[:, :, :], float64[:])',
        'complex128[:, :](complex128[:, :, :], float64[:])'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=True)
def numba_trapz_3dim_array_parallel(y, x):
    result = np.zeros((y.shape[0], y.shape[1]), dtype=type(y[0, 0, 0]))
    for i in prange(y.shape[0]):
        result[i] =__numba_trapz_3dim_array_linear_part(y, x, i)

    return result



@jit(['complex64[:, :, :] (float32[:], float32[:], float32[:], \
                                        complex64[:, :], complex64[:, :], complex64[:, :])',
      'complex128[:, :, :] (float64[:], float64[:], float64[:], \
                                        complex128[:, :], complex128[:, :], complex128[:, :])'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=False)
def numba_3tensordots_1dim_times_2dim(x_float_1dim, y_float_1dim, z_float_1dim, \
                        x_complex_2dim, y_complex_2dim, z_complex_2dim):
    """
    This function can replace snippet

    'foo = np.tensordot(x_float_1dim, x_complex_2dim, axes=0) \n
    foo += np.tensordot(y_float_1dim, y_complex_2dim, axes=0) \n
    foo += np.tensordot(z_float_1dim, z_complex_2dim, axes=0)'

    by

    'foo = get_3_tensordots(x_float_1dim, y_float_1dim, z_float_1dim,
                                    x_complex_2dim, y_complex_2dim, z_complex_2dim)'
    """
    first_shape = x_float_1dim.shape
    second_shape = x_complex_2dim.shape
    result = np.zeros((first_shape[0], second_shape[0], second_shape[1]), dtype=type(x_complex_2dim[0, 0]))
    for i in range(first_shape[0]):
        result[i] = x_float_1dim[i] * x_complex_2dim + \
                        y_float_1dim[i] * y_complex_2dim + \
                            z_float_1dim[i] * z_complex_2dim

    return result


@jit(['complex64[:, :, :] (float32[:], float32[:], float32[:], \
                                        complex64[:, :], complex64[:, :], complex64[:, :])',
      'complex128[:, :, :] (float64[:], float64[:], float64[:], \
                                        complex128[:, :], complex128[:, :], complex128[:, :])'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=True)
def numba_3tensordots_1dim_times_2dim_parallel(x_float_1dim, y_float_1dim, z_float_1dim, \
                        x_complex_2dim, y_complex_2dim, z_complex_2dim):
    first_shape = x_float_1dim.shape
    second_shape = x_complex_2dim.shape
    result = np.zeros((first_shape[0], second_shape[0], second_shape[1]), dtype=type(x_complex_2dim[0, 0]))
    for i in prange(first_shape[0]):
        result[i] = x_float_1dim[i] * x_complex_2dim + \
                        y_float_1dim[i] * y_complex_2dim + \
                            z_float_1dim[i] * z_complex_2dim

    return result


@jit(['types.UniTuple(complex64[:, :, :], 3)(complex64[:, :, :], complex64[:, :, :], complex64[:, :, :], complex64[:, :, :])',
        'types.UniTuple(complex128[:, :, :], 3)(complex128[:, :, :], complex128[:, :, :], complex128[:, :, :], complex128[:, :, :])'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=False)
def evaluate_r_times_eikr(foo_x, foo_y, foo_z, exp_feed):
    '''
    Attention! Sometimes this function can decrease speed on 1 core mode.
    Here foo_x, foo_y, foo_z are supposed to be 2dim arrays with [None, :, :].
    This function can replace snippet

    exp_j = np.exp(1j * exp_feed)
    foo_x_eikr = foo_x * exp_j
    foo_y_eikr = foo_y * exp_j
    foo_z_eikr = foo_z * exp_j

    by

    foo_x_eikr, foo_y_eikr, foo_z_eikr = numba_multiple_on_exp(foo_x, foo_y, foo_z, kr).
    '''
    shape = exp_feed.shape
    result_x = np.zeros((shape[0], shape[1], shape[2]), dtype=type(exp_feed[0, 0, 0]))
    result_y = np.zeros((shape[0], shape[1], shape[2]), dtype=type(exp_feed[0, 0, 0]))
    result_z = np.zeros((shape[0], shape[1], shape[2]), dtype=type(exp_feed[0, 0, 0]))

    for j in range(shape[1]):
        __evaluate_r_times_eikr_linear_part(foo_x, foo_y, foo_z, exp_feed, result_x, result_y, result_z, j, shape[0])

    return result_x, result_y, result_z


@jit(['types.UniTuple(complex64[:, :, :], 3)(complex64[:, :, :], complex64[:, :, :], complex64[:, :, :], complex64[:, :, :])',
        'types.UniTuple(complex128[:, :, :], 3)(complex128[:, :, :], complex128[:, :, :], complex128[:, :, :], complex128[:, :, :])'],
     nopython=True, cache=True, nogil=True, fastmath=True, parallel=True)
def evaluate_r_times_eikr_parallel(foo_x, foo_y, foo_z, exp_feed):
    shape = exp_feed.shape
    result_x = np.zeros((shape[0], shape[1], shape[2]), dtype=type(exp_feed[0, 0, 0]))
    result_y = np.zeros((shape[0], shape[1], shape[2]), dtype=type(exp_feed[0, 0, 0]))
    result_z = np.zeros((shape[0], shape[1], shape[2]), dtype=type(exp_feed[0, 0, 0]))

    for j in prange(shape[1]):
        __evaluate_r_times_eikr_linear_part(foo_x, foo_y, foo_z, exp_feed, result_x, result_y, result_z, j, shape[0])

    return result_x, result_y, result_z


@jit(float64(int64), nopython=True, cache=True, nogil=True)
def jitted_prefactor(m):
    """Returns the prefactor :math:`\sqrt(\frac{(2*m+1)}{2(2m)!} (2m-1)!!`
    without using factorials nor bignum numbers, which makes it jittable.

    Args:
        m (int64): Argument (non-negative)

    Returns:
        :math:`\sqrt(\frac{(2*m+1)!!}{2(2m)!!}`
    """
    res = 1.
    for t in range(2,2*m+2,2):
        res += res/t # equivalent to res *= (t+1)/t, but retains all significant digits
    return (res/2)**0.5


@jit(float64(int64), nopython=True, cache=True, nogil=True)
def prefactor_expansion(m):
    """Expansion of :math:`\sqrt(\frac{(2*m+1)}{2(2m)!} (2m-1)!!` in the limit
    for large :math:`m`.
    The expansion converges very rapidly, so that taking the first 14 terms is
    enough to get every term with double precision down to :math:`m > 10`.
    Further terms are available, but they involve bignum integers.
    Therefore, for terms :math:`m < 10`, it's easier to perform the calculation
    directly.

    Args:
        m (int64): Argument (non-negative)

    Returns:
        :math:`\sqrt(\frac{(2*m+1)!!}{2(2m)!!}`
    """
    if m <= 10:
        return jitted_prefactor(m)
    x = (1/m)**(1/4) / 2
    res = 1/2/x \
        + 3/2*x**3 \
        - 23/4*x**7 \
        + 105/4*x**11 \
        - 1317/16*x**15 \
        + 1053/16*x**19 \
        - 132995/32*x**23 \
        + 3300753/32*x**27 \
        + 24189523/256*x**31 \
        - 7404427407/256*x**35 \
        - 45203760489/512*x**39 \
        + 8818244857071/512*x**43 \
        + 99932439090703/2048*x**47 \
        - 29944926937328991/2048*x**51 \
        # - 173768350561954907/4096*x**55 \
        # + 70443346589090375073/4096*x**59 \
        # + 3299174696912539072131/65536*x**63 \
        # - 1750098951789039471641607/65536*x**67 \
        # - 10306911268683450183973389/131072*x**71 \
        # + 6934593111025074608358220131/131072*x**75 \
        # + 82009190686072900341958975797/524288*x**79 \
        # - 68275536676414175755318233771549/524288*x**83 \
        # - 404866365213072360301912455070149/1048576*x**87 \
        # + 408755654050262791361092019679647223/1048576*x**91 \
        # + 9715746540636352989024348199533962007/8388608*x**95 \
        # - 11697824487409665503661150266687972482035/8388608*x**99
    return res / (np.pi)**(1/4)



"""Provide functionality to store intermediate results in lookup tables (memoize)"""

import pickle
import functools


class Memoize:
    """To be used as a decorator for functions that are memoized."""
    def __init__(self, fn):
        self.fn = fn
        self.memo = {}

    def __call__(self, *args, **kwds):
        if len(self.memo) > 100000:
            self.memo.clear()
        argstr = pickle.dumps(args, 1) + pickle.dumps(kwds, 1)
        if not argstr in self.memo:
            #print("miss") # DEBUG INFO
            self.memo[argstr] = self.fn(*args, **kwds)
        #else:
            #print("hit") # DEBUG INFO
        #print(len(self.memo))
        return self.memo[argstr]
    
    def __get__(self, obj, objtype):
        '''Support instance methods.'''
        return functools.partial(self.__call__, obj)



import numpy as np


def readstl(stlname):
    """
    Reads surface information from STL file
    Args:
        stlname (string): name of STL file

    Returns:
        A list of dictionaries with information about faces of scatterer geometry.
    """
    surfaces = []
    with open(stlname) as f:
        nsurf = 0
        for line in f:
            if "solid" in line and "endsolid" not in line:
                nsurf += 1
                if nsurf > 1:
                    surf = {'nsurf': nsurf - 1,
                            'baris': np.array(baris),
                            'areas': np.array(areas),
                            'normals': np.array(normals)}
                    surfaces.append(surf)
                baricenter = 0
                baris = []
                areas = []
                triangle = []
                normals = []
            if "facet" in line and "end" not in line:
                normal = np.array([float(i) for i in line.split(" ")[2:]])
                normals.append(normal)
                if np.any(np.abs(baricenter) > 0) and bool(triangle):
                    ab = np.array(triangle[1] - triangle[0])
                    ac = np.array(triangle[2] - triangle[0])
                    areas.append(np.linalg.norm(np.cross(ab, ac)) / 2)
                    baris.append(baricenter)
                baricenter = 0
                triangle = []
            if "vertex" in line:
                vertex = np.array([float(i) for i in line.split()[1:]])
                baricenter += vertex / 3
                triangle.append(vertex)

    surf = {'nsurf': nsurf - 1,
            'baris': np.array(baris),
            'areas': np.array(areas),
            'normals': np.array(normals)}
    surfaces.append(surf)
    return surfaces


def writefem(femname,surfaces):
    """
    Writes information about particle geometry to FEM file.
    Args:
        femname (string):    name of FEM file
        surfaces  (list):    information about faces of scatterer geometry
    """    
    fid = open(femname, 'w+')
    lenareas = 0
    print('%7i' % len(surfaces), file=fid)
    ii0 = 1
    for surface in surfaces:
        areas = surface['areas']
        F = surface['normals']
        P = surface['baris']
        print('%7i' % len(areas), file=fid)
        for ii in range(len(areas)):
            print('%7i%17.7E%17.7E%17.7E%17.7E%17.7E%17.7E%17.7E' % (
            ii0 + ii, P[ii, 0], P[ii, 1], P[ii, 2], F[ii, 0], F[ii, 1], F[ii, 2], areas[ii]), file=fid)
        ii0 += len(areas)
    fid.close()


def convert_stl_to_fem(stlname, femname):
    """
    Converts STL to FEM file
    Args:
        stlname (string):    name of STL file
        femname (string):    name of FEM file
    """    
    try:
        surfaces=readstl(stlname)
    except(UnicodeDecodeError):
        raise Exception("Binary STL files are not supported.")
    writefem(femname,surfaces)



import numpy as np
import warnings
import sys

use_gpu = False

default_blocksize = 128


def enable_gpu(enable=True):
    """Sets the use_gpu flag to enable/disable the use of CUDA kernels.

    Args:
        enable (bool): Set use_gpu flag to this value (default=True).
    """
    # todo: do gpuarrays require cuda toolkit? otherwise distinguish if only gpuarrays work but no kernel compilation
    global use_gpu

    if use_gpu == enable:
        return

    if not enable:
        use_gpu = False
        sys.stdout.write('Disabling GPU usage.\n')
        sys.stdout.flush()
        return

    try:
        import pycuda.autoinit
        import pycuda.driver as drv
        from pycuda import gpuarray
        from pycuda.compiler import SourceModule
        import pycuda.cumath
        pycuda_available = True

        current_module = sys.modules[__name__]
        current_module.gpuarray = gpuarray
        current_module.SourceModule = SourceModule
    except ImportError:
        pycuda_available = False

    if not pycuda_available:
        warnings.warn("Unable to import PyCuda - fall back to CPU mode")
        use_gpu = False
        sys.stdout.write('Disabling GPU usage.\n')
        sys.stdout.flush()
    else:
        try:
            test_fun = SourceModule("""
                __global__ void test_kernel(float *x)
                {
                    unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
                    x[i] = 2 * i;
                }""").get_function("test_kernel")
            x = gpuarray.to_gpu(np.zeros(128, dtype=np.float32))
            test_fun(x, block=(128,1,1), grid=(1,1))
            assert np.allclose(x.get(), 2 * np.arange(128))
            use_gpu = True
            sys.stdout.write('Enabling GPU usage.\n')
            sys.stdout.flush()
            return

        except Exception as e:
            warnings.warn("CUDA test kernel failed - fall back to CPU mode")
            print(e)
            use_gpu = False
            sys.stdout.write('Disabling GPU usage.\n')
            sys.stdout.flush()



import sys

from smuthi.version import __version__

try:
    from mpi4py import MPI
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
except:
    mpi_rank = 0


def print_smuthi_header():
    welcome_msg = ("\n" + "*" * 32 + "\n    SMUTHI version " + __version__ + "\n" + "*" * 32 + "\n")
    sys.stdout.write(welcome_msg)
    sys.stdout.flush()


#if mpi_rank == 0:
#    print_smuthi_header()



"""This module contains helper functions for the evaluation of Ewald lattice sums."""

from numba import njit
import numpy as np


@njit()
def reciprocal_lattice_vec(a1, a2):
    """ 
    Args:
        a1 (numpy.ndarray):     lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):     lattice vector 2 in Carthesian coordinates
    Returns:
        2D lattice vectors in reciprocal space, https://en.wikipedia.org/wiki/Reciprocal_lattice
    """ 
    a3 = np.array([0, 0, 1])
    b1 = 2 * np.pi * np.cross(a2, a3) / np.linalg.det(np.concatenate((a1, a2, a3)).reshape(3, 3).astype(np.float64))
    b2 = 2 * np.pi * np.cross(a3, a1) / np.linalg.det(np.concatenate((a2, a3, a1)).reshape(3, 3).astype(np.float64))    
    return b1[:2], b2[:2]


@njit()
def n1n2_indices(x):
    """ Indices n1, n2 that specify lattice point R = n1*a1 + n2*a2 of unit cells in a distance x around
        the arbitrary central unit cell.   
    Args:
        x (int):                distance x from the arbitrary center unit cell
    Returns:
        n1 (numpy.ndarray):     index of lattice vector a1
        n2 (numpy.ndarray):     index of lattice vector a2
    """
    if x == 0:
        return np.array([0], np.float64), np.array([0], np.float64)
    
    array = np.zeros((2, 8*x), np.float64)   
    idx = 0
    for n1 in [-x, x]:
        for n2 in np.arange(-x, x+1): 
            array[:, idx] = [n1, n2]
            idx += 1
    for n1 in np.arange(-x+1, x):
        for n2 in [-x, x]:
            array[:, idx] = [n1, n2]
            idx += 1
    return array[0, :], array[1, :]


# factorials up to 20!
factorial_table = np.array([1, 1, 2, 6, 24, 120, 720, 5040, 40320,
                   362880, 3628800, 39916800, 479001600,
                   6227020800, 87178291200, 1307674368000,
                   20922789888000, 355687428096000, 6402373705728000,
                   121645100408832000, 2432902008176640000], np.int64)
# x! / 20! up to x = 30
factorial_table_2 = np.array([21, 462, 10626, 255024, 6375600, 165765600,
                              4475671200, 125318793600, 3634245014400,
                              109027350432000], np.int64)

@njit()
def numba_factorial(x):
    """ Returns factorials for x <= 30.
    Args:
        x (numpy.ndarray):      argument array of type np.int64
    Returns:
        res (numpy.ndarray):    2 dimensional array that splits the factorial into
                                two factors x! = x2! * x1! to allow storage as np.int64
                                First, for arguments 20 < x2 <= 30, 21*...*x2
                                Second, for arguments x1 <= 20, 1*2*...*x1
    """
    res = np.zeros((x.shape[0], 2), np.int64)
    for idx, xx in enumerate(x):
        if xx > 30:
            raise ValueError('Only factorials of x <= 30 available!')
        elif xx > 20:
            res[idx, 0] = factorial_table_2[xx - 21]
            res[idx, 1] = factorial_table[20]
        else:
            res[idx, 0] = 1
            res[idx, 1] = factorial_table[xx]
    return res


double_factorial_table = np.array([1, 1, 2, 3, 8, 15, 48, 105, 384, 945, 3840,
                                   10395, 46080, 135135, 645120, 2027025,
                                   10321920, 34459425, 185794560, 654729075,
                                   3715891200, 13749310575, 81749606400,
                                   316234143225, 1961990553600, 7905853580625,
                                   51011754393600, 213458046676875,
                                   1428329123020800, 6190283353629375, 
                                   42849873690624000], np.int64)

@njit()
def numba_double_factorial(x):
    """ Returns double factorials for x <= 30.
    Args:
        x (numpy.ndarray):      argument array of type np.int64
    """
    if x > 30:
        raise ValueError
    return double_factorial_table[x]


###############################################################################
from numba.extending import get_cython_function_address
import ctypes
_PTR = ctypes.POINTER
_dble = ctypes.c_double
_ptr_dble = _PTR(_dble)


def load_wofz(name):
    try:
        addr = get_cython_function_address(name, "wofz")
    except: return None
    else: return addr


import_wofz_possible_file_paths = ["cython_speedups",
         "smuthi.utility.cython.cython_speedups",
         "utility.cython.cython_speedups"]


def raise_wofz_exception():
    raise Exception('''
                    In smuthi/periodic/ewald_helper.py Numba needs to know the location of the function "wofz".\n
                    This function is imported from the Cython file in smuthi/utility/cython/cython_speedups_api.h.\n
                    We tried three main methods to import this file, but all three failed. :( \n
                    This likely means that you are trying to run Smuthi for a non-standard directory and \n
                    Smuthi is unable to deterine where to find the cython_speedups file. If you need to run \n
                    Smuthi in a non-standard directory, you can add the proper file paths to the import_wofz_possible_file_paths\n
                    variable on line 118 in the ewald_helper file. 
                    
                    Alternatively, this error may occur because cython_speedups_api.h does not exist and needs to be\n
                    recompiled from cython_speedups.pyx. If this is the case, please follow the instructions in the file\n
                    smuthi/utility/cython/compile_speedups_from_c.py or, if compile_speedups.c does not exist \n
                    follow the instructions in compile_speedups_from_pyx.py. 
                    ''')


try: 
    for name in import_wofz_possible_file_paths:
        addr = load_wofz(name)
        if addr != None:
            break    
except:
    raise_wofz_exception()

if addr is None:
    raise_wofz_exception()

functype = ctypes.CFUNCTYPE(None, _dble, _dble, _ptr_dble, _ptr_dble)
wofz_fn = functype(addr)


@njit()
def numba_wofz_complex(x):
    """ Numba compatible version of the Faddevva function - replacing scipy.special.wofz().
        The cwrapper is copied from: https://github.com/numba/numba/issues/3086
    """
    out_real = np.empty(1,dtype=np.float64)
    out_imag = np.empty(1,dtype=np.float64)
    wofz_fn(np.real(x), np.imag(x), out_real.ctypes, out_imag.ctypes)    
    return np.complex(out_real[0] + 1j * out_imag[0])
###############################################################################


@njit()
def upper_incomplete_gamma_fun(n, x_vec):
    """ Recursion formula of the upper incomplete Gamma function.  
        Kambe, Z. Naturforschg. 22a, 322-330, 1967. (42) und Appendix 2 (A8), (A9)
    Args:
        n (int):                    order
        x_vec (numpy.ndarray):      argument array
    Returns:
        gamma_vec (numpy.ndarray):  values of the upper incomplete gamma function for 0 ... n 
                                    dimension: len(x_vec) times (n+1)          
    """   
    gamma_vec = np.zeros((x_vec.shape[0], n+1), np.complex128)
    for idx_x, x in enumerate(x_vec):         
        arg_x = np.arctan2(x.imag, x.real)
        if abs(arg_x - np.pi) < 1e-08:
            arg_x = -arg_x
            val = np.sqrt(np.abs(x)) * np.exp(0.5j * (arg_x + np.pi))
            w = numba_wofz_complex(val)
            gamma_vec[idx_x, 0] = np.sqrt(np.pi) * np.exp(-x) * w 
            for idx_n in range(1, n+1):
                b = 0.5 - idx_n
                gamma_vec[idx_x, idx_n] = (gamma_vec[idx_x, idx_n-1] + x ** b * np.exp(-x)) / b
        else:
            val = np.sqrt(np.abs(x)) * np.exp(0.5j * (arg_x + np.pi))
            w = numba_wofz_complex(val)
            gamma_vec[idx_x, 0] = np.sqrt(np.pi) * np.exp(-x) * w
            for idx_n in range(1, n+1):
                b = 0.5 - idx_n
                gamma_vec[idx_x, idx_n] = (gamma_vec[idx_x, idx_n-1] - x ** b * np.exp(-x)) / b
    return gamma_vec


@njit()
def int_recursion(L, eta, k, R_vec):
    """ Integral recursion for the Ewald sum's real space summand 
        Kambe, Z. Naturforschg. 22a, 322-330, 1967. Appendix 2, (A11) - (A14)
    Args:
        L (int):                multipole degree
        eta (float):            Ewald sum separation parameter 
        k (complex):            initial field's wavenumber
        R_vec (numpy.ndarray):  real space unit cell displacement vector     
    Returns:
        Ewald sum's real space summand integral 
    """
    shape = R_vec.shape[0]
    alpha = k ** 2 / (4 * eta ** 2)
    I = np.zeros((shape, L+2), np.complex128)
    w = np.zeros(shape, np.complex128)
    for idx in range(shape):
        w[idx] = numba_wofz_complex(np.sqrt(alpha) + 1j * k * R_vec[idx] / (2 * np.sqrt(alpha)))
    exp_term =  np.exp(alpha - (k * R_vec) ** 2 / (4 * alpha))
    
    I[:, 0] = np.sqrt(np.pi) * exp_term * w.imag 
    I[:, 1] = np.sqrt(np.pi) * 2 / (k * R_vec) * exp_term * w.real
    
    for idx in range(2, L+2):
        I[:, idx] = (2 / (k * R_vec)) ** 2 * ((2 * (idx-2) + 1) / 2 * I[:, idx-1] - I[:, idx-2] + alpha ** (-(idx-2) - 0.5) * exp_term)        
    return I[:, -1]


@njit()
def delta_n(n_max, gamma, cz, eta):
    """ Integral recursion for the Ewald sum's reciprocal space summand to account for the
        coupling between one particle (S_0i) and a different particle's (S_0j) periodic arrangement.
        Kambe, Z. Naturforschg. 23a, 1280-1294, 1968. Appendix 3, (A3.1) - (A.3.7)
    Args:
        n_max (int):            maximal order
        gamma (numpy.ndarray):  sqrt(k ** 2 - kgt ** 2)
        cz (float):             particle displacement in along z-axis
        eta (float):            Ewald sum separation parameter 
    Returns:
        Ewald sum's reciprocal space summand intergal
    """
    shape = gamma.shape[0]
    I = np.zeros((shape, n_max + 1), np.complex128)
    w_mn = np.zeros(shape, np.complex128)
    w_pl = np.zeros(shape, np.complex128)
    
    x = -gamma ** 2 / (4 * eta ** 2)
    rootx = np.zeros(gamma.shape[0], np.complex128)
    z = np.zeros(gamma.shape[0], np.complex128)     
    
    sign = np.where(x.real < 0, 1.0, 0.0)
    rootx = -1j * (sign * np.abs(x)) ** 0.5 + ((1 - sign) * x) ** 0.5
    z = sign * (gamma * cz) + (1 - sign) * 1j * np.abs(gamma * cz)
    
    zsqr = (gamma * cz) ** 2
    exp_term = np.exp(-x + zsqr / (4 * x))
    for idx in range(shape):
        w_mn[idx] = numba_wofz_complex(-z[idx] / (2 * rootx[idx]) + 1j * rootx[idx])
        w_pl[idx] = numba_wofz_complex(z[idx] / (2 * rootx[idx]) + 1j * rootx[idx])
            
    I[:, 0] = np.pi ** 0.5 / 2 * exp_term * (w_mn + w_pl)
    if n_max == 0:
        return I
    I[:, 1] = 1j * np.pi ** 0.5 / z * exp_term * (w_mn - w_pl)
    
    for idx in range(2, n_max + 1):
        I[:, idx] = 4 / zsqr * ((1.5 - idx) * I[:, idx-1] - I[:, idx-2] + rootx * x ** (1 - idx) * exp_term) 
    return I
        
        
@njit()
def normalization_spherical_harmonics(M):  
    """ Normalization factor between the definition of spherical harmonics in Smuthi
        (following Doicu, Light Scattering by Systems of Particles, Berlin, 2006.) and the definitions
        in Kambe and Beutel (following Tsang, Theory of Microwave Remote Sensing, 1985.)
        for details see: https://doi.org/10.5445/IR/1000141626 (Appendix C, p.135).
    """
    if M >= 0:
        return np.sqrt(2 * np.pi) * (-1) ** (-M)
    else:
        return np.sqrt(2 * np.pi)


"""This subpackage contains functionality that has to do with light scattering
in periodic particle arrangements.
The __init__ module contains some helper functions (e.g. with respect to
periodic particle arrangements) and is the place to store the default separation
parameter that splits the Ewald lattice sum into real and reciprocal space."""

import numpy as np



"""Default Ewald sum separation parameter eta - needs to be set, e.g. at beginning of a periodic simulation"""
default_Ewald_sum_separation = None


def set_ewald_sum_separation(a1, a2, initial_field=None, particle_list=None, layer_system=None, magM=None):
    """ Set the separation parameter eta that splits the evaluation of the Ewald lattice sum
        into real and reciprocal space.
    Args:
        a1 (numpy.ndarray):                                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                                 lattice vector 2 in Carthesian coordinates
        initial_field (smuthi.initial_field.PlaneWave):     initial plane wave object
        layer_system (smuthi.layers.LayerSystem):           stratified medium
        particle_list (list):                               list of smuthi.particles.Particle objects
        magM (float):                                       maximum tolerated magnitude with regard to
                                                            target accuracy
    Returns:
        Separation parameter eta 
    """
    if magM == None or magM == 1:
        return np.sqrt(np.pi / np.linalg.norm(np.cross(a1, a2)))
    else:
        z_list = [particle.position[2] for particle in particle_list]
        is_list = [layer_system.layer_number(z) for z in z_list]
        assert is_list.count(is_list[0]) == len(is_list)  # all particles in same layer?
        
        k = complex(2 * np.pi * layer_system.refractive_indices[is_list[0]] / initial_field.vacuum_wavelength)
        
        if initial_field.polar_angle < np.pi:
            pwe_exc = initial_field.plane_wave_expansion(layer_system, is_list[0])[0]
        else:
            pwe_exc = initial_field.plane_wave_expansion(layer_system, is_list[0])[1]
        k0t = np.array([pwe_exc.k_parallel[0] * np.cos(pwe_exc.azimuthal_angles)[0],
                        pwe_exc.k_parallel[0] * np.sin(pwe_exc.azimuthal_angles)[0]])     
        
        return (np.sqrt(k ** 2 - np.linalg.norm(k0t) ** 2) / (2 * np.log(magM))).real



"""This module contains functions to compute Ewald lattice sums."""

import smuthi.periodicboundaries.ewald_helper as pbeh
import smuthi.utility.math as sf
from numba import njit
import numpy as np



@njit()
def D_LM(L, M, k, k0t, a1, a2, eta, c=np.zeros(3, np.float64)):
    """ Computation of the Ewald sum to account for the coupling between one particle
        and a periodic particle arrangement.
    Args:
        L (int):                multipole degree
        M (int):                multipole order
        k (complex):            wavenumber
        k0t (numpy.ndarray):    complex in-plane wave vector in Carthesian coordinates 
        a1 (numpy.ndarray):     lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):     lattice vector 2 in Carthesian coordinates      
        eta (float):            Ewald sum separation parameter
        c (numpy.ndarray):      displacment vector between emitting and receiving particle
    Returns:
        D_LM (complex):         Ewald sum D1_LM + D2_LM + D3_00
    """        
    A = np.linalg.norm(np.cross(a1, a2))
    
    if np.linalg.norm(c) == 0:
        D1_LM_val = D1_LM(L, M, k, k0t, a1, a2, eta, A)
        D2_LM_val = D2_LM(L, M, k, k0t, a1, a2, eta)                
        if L == 0:
            D3_00_val = D3_00(k, eta)
            return D1_LM_val + D2_LM_val + D3_00_val
        else:
            return D1_LM_val + D2_LM_val
    else:
        # distance from receiving to emitting particle (opposing to the common distance vector)
        D1_LM_val = D1_LM_ij(L, M, -c, k, k0t, a1, a2, eta, A)
        D2_LM_val = D2_LM_ij(L, M, -c, k, k0t, a1, a2, eta)
        return D1_LM_val + D2_LM_val  
    
    
@njit()
def D1_LM(L, M, k, k0t, a1, a2, eta, A, c=np.zeros(3)):
    """ Ewald sum's reciprocal space summand to account for the coupling between one particle
        with it's own periodic arrangement. Beutel, arXiv:2004.08098,  Appendix D, (D2a).
    Args:
        L (int):                multipole degree
        M (int):                multipole order
        k (complex):            wavenumber
        k0t (numpy.ndarray):    complex in-plane wave vector in Carthesian coordinates 
        a1 (numpy.ndarray):     lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):     lattice vector 2 in Carthesian coordinates
        eta (float):            Ewald sum separation parameter 
        A (float):              area of the unit cell
        c (numpy.ndarray):      displacement vector betwenn particle S0j and S0j' in Carthesian coordinates
    Returns:
        D1_LM (complex):        reciprocal space summand of Ewald sum D_LM
    """
    if (L - abs(M)) % 2:
        return 0
        
    b1, b2 = pbeh.reciprocal_lattice_vec(a1, a2)      
    root = (2 * L + 1) ** 0.5  \
            * pbeh.numba_factorial(np.array([L - M], np.int32))[0, 0] ** 0.5 * pbeh.numba_factorial(np.array([L - M], np.int32))[0, 1] ** 0.5 \
            * pbeh.numba_factorial(np.array([L + M], np.int32))[0, 0] ** 0.5 * pbeh.numba_factorial(np.array([L + M], np.int32))[0, 1] ** 0.5
    prefac =  (1j) ** M * root / (A * k * (2 * k) ** L)
    
    x=0        
    n1, n2 = pbeh.n1n2_indices(x)
    D1_LM = compute_sum_D1LM(L, M, k, k0t, n1, n2, b1, b2, eta, c)
    
    flag = 0
    while flag == 0:
        D1_LM_prime = D1_LM        
        x += 1
        n1, n2 = pbeh.n1n2_indices(x)
        D1_LM += compute_sum_D1LM(L, M, k, k0t, n1, n2, b1, b2, eta, c)  
        if not D1_LM_prime == complex(0):
            if np.linalg.norm(np.array([D1_LM - D1_LM_prime])) / np.linalg.norm(np.array([D1_LM_prime])) < 1e-5:
                flag = 1
    
    return prefac * D1_LM


@njit()
def compute_sum_D1LM(L, M, k, k0t, n1, n2, b1, b2, eta, c=np.zeros(3)):
    """ Helper function for D1_LM(). """    
    shape = n1.shape[0]    
    gamma_g = np.zeros(shape, np.complex128)
    kgt_r = np.zeros(shape, np.float64)
    kgt_phi = np.zeros(shape, np.float64)
    
    Bgt = n1.reshape(shape, 1) * b1.reshape(1, 2) + n2.reshape(shape, 1) * b2.reshape(1, 2)
    kgt = k0t + Bgt
    for idx in range(shape):
        kgt_r[idx] = np.linalg.norm(kgt[idx, :])
        kgt_phi[idx] = np.arctan2(kgt[idx, 1], kgt[idx, 0])
    
    gamma_g = np.sqrt(k ** 2 - kgt_r ** 2)  
    gamma_g[np.where(gamma_g == complex(0))[0]] += 1e-10j                 
    eiMphi = np.exp(1j * M * kgt_phi)
    emnikgtc = np.exp(-1j * np.dot(kgt, c[:2])) 
    
    n = np.arange(int((L - abs(M)) / 2) + 1)
    denom = pbeh.numba_factorial(n)[:, 0] * pbeh.numba_factorial(n)[:, 1] \
            * pbeh.numba_factorial(int((L + M) / 2) - n)[:, 0] * pbeh.numba_factorial(int((L + M) / 2) - n)[:, 1] \
            * pbeh.numba_factorial(int((L - M) / 2) - n)[:, 0] * pbeh.numba_factorial(int((L - M) / 2) - n)[:, 1]
    nom = gamma_g.reshape(shape, 1) ** (2 * n - 1).reshape(1, n.shape[0]) * kgt_r.reshape(shape, 1) ** (L - 2 * n).reshape(1, n.shape[0])
    gamma_fun = pbeh.upper_incomplete_gamma_fun(n[-1], -gamma_g ** 2 / (4 * eta ** 2))  
    return np.sum(emnikgtc * eiMphi * np.sum(gamma_fun * nom / denom, axis=1))


@njit()
def D2_LM(L, M, k, k0t, a1, a2, eta):
    """ Ewald sum's real space summand to account for the coupling between one particle
        with it's own periodic arrangement. Beutel, arXiv:2004.08098,  Appendix D, (D2b).
    Args:
        L (int):                multipole degree
        M (int):                multipole order
        k (complex):            wavenumber
        k0t (numpy.ndarray):    complex in-plane wave vector in Carthesian coordinates 
        a1 (numpy.ndarray):     lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):     lattice vector 2 in Carthesian coordinates
        eta (float):            Ewald sum separation parameter      
    Returns:
        D2_LM (complex):        real space summand of Ewald sum D_LM
    """
    if (L - abs(M)) % 2:
        return 0
    
    frac = -1j * (-1) ** ((L + M) / 2) / (2 ** (L + 1) * np.pi \
            * pbeh.numba_factorial(np.array([(L - M) / 2], np.int32))[0, 0] * pbeh.numba_factorial(np.array([(L - M) / 2], np.int32))[0, 1] \
            * pbeh.numba_factorial(np.array([(L + M) / 2], np.int32))[0, 0] * pbeh.numba_factorial(np.array([(L + M) / 2], np.int32))[0, 1]) 
    root = (2 * L + 1) ** 0.5  \
            * pbeh.numba_factorial(np.array([L - M], np.int32))[0, 0] ** 0.5 * pbeh.numba_factorial(np.array([L - M], np.int32))[0, 1] ** 0.5 \
            * pbeh.numba_factorial(np.array([L + M], np.int32))[0, 0] ** 0.5 * pbeh.numba_factorial(np.array([L + M], np.int32))[0, 1] ** 0.5
    prefac = frac * root
      
    x=1
    n1, n2 = pbeh.n1n2_indices(x)
    D2_LM = compute_sum_D2LM(L, M, k, k0t, n1, n2, a1, a2, eta)
    
    flag = 0
    while flag == 0:
        D2_LM_prime = D2_LM       
        x += 1
        n1, n2 = pbeh.n1n2_indices(x)
        D2_LM += compute_sum_D2LM(L, M, k, k0t, n1, n2, a1, a2, eta) 
        if not D2_LM_prime == complex(0):
            if np.linalg.norm(np.array([D2_LM - D2_LM_prime])) / np.linalg.norm(np.array([D2_LM_prime])) < 1e-5:
                flag = 1
       
    return prefac * D2_LM


@njit()
def compute_sum_D2LM(L, M, k, k0t, n1, n2, a1, a2, eta):
    """ Helper function for D2_LM(). """ 
    shape = n1.shape[0] 
    Rn_r = np.zeros(shape, np.float64)
    Rn_phi = np.zeros(shape, np.float64)
    
    Rn = n1.reshape(shape, 1) * a1[:2].reshape(1, 2) + n2.reshape(shape, 1) * a2[:2].reshape(1, 2)
    for idx in range(shape):
        Rn_r[idx] = np.linalg.norm(Rn[idx, :])
        Rn_phi[idx] = np.arctan2(Rn[idx, 1], Rn[idx, 0])       
    k0tRn = np.dot(k0t, np.transpose(Rn))

    int_val = (k ** 2 / 4) ** (L + 0.5) * pbeh.int_recursion(L, eta, k, Rn_r) 
    return  np.sum(np.exp(1j * (k0tRn + M * (Rn_phi + np.pi))) / k * (2 * Rn_r / k) ** L * int_val)


@njit()
def D3_00(k, eta):
    """ Central lattice point correction for the Ewald sum to account for the coupling between
        one particle and it's own periodic arrangement. Beutel, arXiv:2004.08098,  Appendix D, (D2c).
    Args:
        k (complex):        wavenumber
        eta (float):        Ewald sum separation parameter      
    Returns:
        D3_00 (complex):    central lattice point correction of Ewald sum D_LM
    """
    x = np.array([-k ** 2 / (4 * eta ** 2)])
    return 1 / (4 * np.pi) * pbeh.upper_incomplete_gamma_fun(1, x)[0, -1]


@njit()
def D1_LM_ij(L, M, c, k, k0t, a1, a2, eta, A):
    """ Ewald sum's reciprocal space summand to account for the coupling between one particle (S_0i)
        and a different particle's (S_0j) periodic arrangement. Kambe, Z. Naturforschung 23a, (3.17).
    Args:        
        L (int):                multipole degree
        M (int):                multipole order
        c (numpy.ndarray):      displacement vector betwenn particle S0j and S0j' in Carthesian coordinates
        k (complex):            wavenumber
        k0t (numpy.ndarray):    complex in-plane wave vector in Carthesian coordinates 
        a1 (numpy.ndarray):     lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):     lattice vector 2 in Carthesian coordinates
        eta (float):            Ewald sum separation parameter  
        A (float):              area of one unit cell
    Returns:
        D1LM (complex):        reciprocal space summand of Ewald sum D_LM
    """   
    if c[2] == 0:
        return D1_LM(L, M, k, k0t, a1, a2, eta, A, c)
    else:  
        b1, b2 = pbeh.reciprocal_lattice_vec(a1, a2)
        root = (2 * L + 1) ** 0.5  \
            * pbeh.numba_factorial(np.array([L - M], np.int32))[0, 0] ** 0.5 * pbeh.numba_factorial(np.array([L - M], np.int32))[0, 1] ** 0.5 \
            * pbeh.numba_factorial(np.array([L + M], np.int32))[0, 0] ** 0.5 * pbeh.numba_factorial(np.array([L + M], np.int32))[0, 1] ** 0.5
        prefac = (-1j) ** M * root / ((-2) ** L * A * k ** 2)
        
        x=0        
        n1, n2 = pbeh.n1n2_indices(x)
        D1LM = compute_sum_D1LM_ij(L, M, k, k0t, n1, n2, b1, b2, eta, A, c)
                
        flag = 0
        while flag == 0:
            D1_LM_prime = D1LM        
            x += 1
            n1, n2 = pbeh.n1n2_indices(x)
            D1LM += compute_sum_D1LM_ij(L, M, k, k0t, n1, n2, b1, b2, eta, A, c)  
            if not D1_LM_prime == complex(0):
                if np.linalg.norm(np.array([D1LM - D1_LM_prime])) / np.linalg.norm(np.array([D1_LM_prime])) < 1e-5:
                    flag = 1
        return prefac * D1LM
    
    
@njit()
def compute_sum_D1LM_ij(L, M, k, k0t, n1, n2, b1, b2, eta, A, c):
    """ Helper function for D1_LM_ij(). """ 
    shape = n1.shape[0]
    gamma_g = np.zeros(shape, np.complex128)
    kgt_r = np.zeros(shape, np.float64)
    kgt_phi = np.zeros(shape, np.float64)
    
    Bgt = n1.reshape(shape, 1) * b1.reshape(1, 2) + n2.reshape(shape, 1) * b2.reshape(1, 2)
    kgt = k0t + Bgt
    for idx in range(shape):
        kgt_r[idx] = np.linalg.norm(kgt[idx, :])
        kgt_phi[idx] = np.arctan2(kgt[idx, 1], kgt[idx, 0])
    
    gamma_g = np.sqrt(k ** 2 - kgt_r ** 2)      
    gamma_g[np.where(gamma_g == complex(0))[0]] += 1e-10j            
    eiMphi = np.exp(1j * M * kgt_phi)
    emnikgtc = np.exp(-1j * np.dot(kgt, c[:2]))   
    
    n = np.arange(0, L - abs(M) + 1)
    inner_sum = np.zeros((shape, n.shape[0]), np.complex128)
    for nn in n:    
        s = np.arange(nn, min(L - abs(M), 2 * nn) + 1)
        if not (L - abs(M)) % 2:
            s = s[np.argwhere(s % 2 == 0).flatten()]
        else:
            s = s[np.argwhere(s % 2).flatten()]
            
        if not s.shape[0] == 0:    
            denom = pbeh.numba_factorial(2 * nn - s)[:, 0] * pbeh.numba_factorial(2 * nn - s)[:, 1] \
                    * pbeh.numba_factorial(s - nn)[:, 0] * pbeh.numba_factorial(s - nn)[:, 1] \
                    * pbeh.numba_factorial(np.asarray((L + abs(M) - s) / 2, np.int32))[:, 0] \
                    * pbeh.numba_factorial(np.asarray((L + abs(M) - s) / 2, np.int32))[:, 1] \
                    * pbeh.numba_factorial(np.asarray((L - abs(M) - s) / 2, np.int32))[:, 0] \
                    * pbeh.numba_factorial(np.asarray((L - abs(M) - s) / 2, np.int32))[:, 1]
            nom = (-k * c[2]) ** (2 * nn - s) * (kgt_r / k).reshape(shape, 1) ** (L - s)
            inner_sum[:, nn] = np.sum(nom / denom, axis=1)
            
    delta = pbeh.delta_n(n[-1], gamma_g, c[2], eta)
    return np.sum(emnikgtc * eiMphi * np.sum((gamma_g / k).reshape(shape, 1) ** (2 * n - 1) * delta * inner_sum, axis=1))    


@njit()
def D2_LM_ij(L, M, c, k, k0t, a1, a2, eta):
    """ Ewald sum's real space summand to account for the coupling between one particle (S_0i)
        and a different particle's (S_0j) periodic arrangement. Kambe, Z. Naturforschung 22a, (46b).
    Args:
        L (int):                multipole degree
        M (int):                multipole order
        c (numpy.ndarray):      displacement vector betwenn particle S0j and S0j' in Carthesian coordinates
        k (complex):            wavenumber
        k0t (numpy.ndarray):    complex in-plane wave vector in Carthesian coordinates 
        a1 (numpy.ndarray):     lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):     lattice vector 2 in Carthesian coordinates
        eta (float):            Ewald sum separation parameter 
    Returns:
        D2_LM (complex):        real space summand of Ewald sum D_LM
    """
    if c[2] == 0 and (L - abs(M)) % 2:
        return 0
  
    prefac = -1j * np.sqrt(2 / np.pi)
      
    x=0
    n1, n2 = pbeh.n1n2_indices(x)
    D2_LM = compute_sum_D2LM_ij(L, M, k, k0t, n1, n2, a1, a2, eta, c)
    
    flag = 0
    while flag == 0:
        D2_LM_prime = D2_LM       
        x += 1
        n1, n2 = pbeh.n1n2_indices(x)
        D2_LM += compute_sum_D2LM_ij(L, M, k, k0t, n1, n2, a1, a2, eta, c)       
        if not D2_LM_prime == complex(0):
            if np.linalg.norm(np.array([D2_LM - D2_LM_prime])) / np.linalg.norm(np.array([D2_LM_prime])) < 1e-5:
                flag = 1
    return prefac * D2_LM


@njit()
def compute_sum_D2LM_ij(L, M, k, k0t, n1, n2, a1, a2, eta, c):
    """ Helper function for D2_LM_jjp(). """ 
    shape = n1.shape[0]
    mnRnc_r = np.zeros(shape, np.float64)
    mnRnc_theta = np.zeros(shape, np.float64)
    mnRnc_phi = np.zeros(shape, np.float64)
    
    Rn = n1.reshape(shape, 1) * a1[:2].reshape(1, 2) + n2.reshape(shape, 1) * a2[:2].reshape(1, 2)
    mnRnc = -(n1.reshape(shape, 1) * a1.reshape(1, 3) + n2.reshape(shape, 1) * a2.reshape(1, 3) + c)
    for idx in range(shape):
        mnRnc_r[idx] = np.linalg.norm(mnRnc[idx, :])
        mnRnc_theta[idx] = np.arctan2(np.linalg.norm(mnRnc[idx, :2]), mnRnc[idx, 2])
        mnRnc_phi[idx] = np.arctan2(mnRnc[idx, 1], mnRnc[idx, 0])
             
    eik0tRn = np.exp(1j * np.dot(k0t, np.transpose(Rn)))
    Y_LM = sf.legendre_normalized_numbed(np.cos(mnRnc_theta), np.sin(mnRnc_theta),
                               max(1, L))[0][L, abs(M)] * np.exp(1j * M * mnRnc_phi) / pbeh.normalization_spherical_harmonics(M)
    
    int_val = (1 / 2) ** (L + 1.5) * pbeh.int_recursion(L, eta, k, mnRnc_r)
    return  np.sum(eik0tRn * (k * mnRnc_r) ** L * Y_LM * int_val)